# -*- coding: utf-8 -*-
"""
string normalizer services module.
"""

from pyrin.utilities.string.normalizer import StringNormalizerPackage
from pyrin.application.services import get_component


def normalize(value, *normalizers, **options):
    """
    normalizes the given value.

    :param str value: value to be normalized.

    :param str normalizers: normalizer names to be used.
                            they will be used in the order of their appearance.
                            if not provided, all normalizers will be used with
                            the order of their priority attribute.

    :keyword list[str] filters: list of items to be removed from string.
                                defaults to None. it will only be used
                                for `filter` normalizer.

    :keyword bool ignore_case: remove `filters` from string in case-insensitive
                               way. defaults to True if not provided.

    :keyword bool strip: strip spaces from both ends of string on each
                         normalization step. defaults to True if not provided.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :raises StringNormalizerDoesNotExistError: string normalizer does not exist error.

    :returns: normalized value.
    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).normalize(value,
                                                                           *normalizers,
                                                                           **options)


def register_normalizer(instance, **options):
    """
    registers a new string normalizer or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a normalizer which is already registered.

    :param AbstractStringNormalizerBase instance: normalizer to be registered.
                                                  it must be an instance of
                                                  AbstractStringNormalizerBase.

    :keyword bool replace: specifies that if there is another registered
                           normalizer with the same name, replace it with
                           the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidStringNormalizerTypeError: invalid string normalizer type error.
    :raises DuplicatedStringNormalizerError: duplicated string normalizer error.
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).register_normalizer(instance,
                                                                                     **options)


def get_normalizer(name, **options):
    """
    gets the string normalizer with given name.

    it raises an error if normalizer does not exist.

    :raises StringNormalizerDoesNotExistError: string normalizer does not exist error.

    :rtype: AbstractStringNormalizerBase
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).get_normalizer(name, **options)


def space(value, **options):
    """
    remove all spaces from the string.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).space(value, **options)


def duplicate_space(value, **options):
    """
    replace all duplicate spaces with a single one.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).duplicate_space(value,
                                                                                 **options)


def lowercase(value, **options):
    """
    lowercase the string.

    :param str value: value to be lower-cased.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).lowercase(value, **options)


def uppercase(value, **options):
    """
    uppercase the string.

    :param str value: value to be upper-cased.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).uppercase(value, **options)


def title_case(value, **options):
    """
    title case the string.

    :param str value: value to be title-cased.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).title_case(value, **options)


def filter(value, **options):
    """
    remove filters from the string.

    :param str value: value to be filtered.

    :keyword list[str] filters: list of items to be removed from value.

    :keyword bool ignore_case: remove `filters` from string in case-insensitive
                               way. defaults to True if not provided.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).filter(value, **options)


def persian_sign(value, **options):
    """
    remove common persian signs from the string.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).persian_sign(value, **options)


def latin_sign(value, **options):
    """
    remove common latin signs from the string.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).latin_sign(value, **options)


def persian_number(value, **options):
    """
    replace persian numbers with latin numbers.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).persian_number(value, **options)


def arabic_number(value, **options):
    """
    replace arabic numbers with latin numbers.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).arabic_number(value, **options)


def persian_letter(value, **options):
    """
    normalize persian letters.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).persian_letter(value, **options)


def latin_letter(value, **options):
    """
    normalize latin letters.

    :param str value: value to be normalized.

    :keyword bool normalize_none: specifies that if given value is None,
                                  return empty string instead of None.
                                  defaults to False if not provided.

    :rtype: str
    """

    return get_component(StringNormalizerPackage.COMPONENT_NAME).latin_letter(value, **options)



# -*- coding: utf-8 -*-
"""
response wrappers structs module.
"""

from pyrin.core.structs import Context


class ResponseContext(Context):
    """
    context class to hold response contextual data.
    """

    attribute_error_message = 'Property [{name}] not found in response context.'



# -*- coding: utf-8 -*-
"""
datetime manager module.
"""

from datetime import datetime

import pytz

import pyrin.configuration.services as config_services
import pyrin.security.session.services as session_services
import pyrin.utils.datetime as datetime_utils

from pyrin.core.structs import Manager
from pyrin.globalization.datetime import DateTimePackage


class DateTimeManager(Manager):
    """
    datetime manager class.
    """

    package_class = DateTimePackage

    def __init__(self, **options):
        """
        initializes an instance of DateTimeManager.
        """

        super().__init__()

        server_timezone = config_services.get('globalization',
                                              'locale', 'babel_default_timezone')

        client_timezone = config_services.get('globalization',
                                              'locale', 'client_timezone')

        self._server_timezone = pytz.timezone(server_timezone)
        self._client_timezone = pytz.timezone(client_timezone)

    def _add_timezone(self, value, server):
        """
        adds the current timezone info into input value if it has no timezone info.

        :param datetime value: value to add timezone info into it.
        :param bool server: specifies that server or client timezone must be added.

        :rtype: datetime
        """

        localized_value = value
        if value.tzinfo is None:
            localized_value = self.localize(value, server=server)

        return localized_value

    def now(self, server=True, timezone=None):
        """
        gets the current datetime based on requested timezone.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get datetime based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :rtype: datetime
        """

        if timezone not in (None, ''):
            timezone = self.get_timezone(timezone)
        else:
            timezone = self.get_current_timezone(server=server)

        return datetime.now(timezone)

    def client_now(self):
        """
        gets the current datetime based on client timezone.

        this is a helper method to let get the client datetime
        without providing value to `now` method.

        :rtype: datetime
        """

        return self.now(server=False)

    def current_year(self, server=True, timezone=None):
        """
        gets the current year based on requested timezone.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get year based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :rtype: int
        """

        now = self.now(server=server, timezone=timezone)
        return now.year

    def current_client_year(self):
        """
        gets the current year based on client timezone.

        this is a helper method to let get the client year
        without providing value to `current_year` method.

        :rtype: int
        """

        client_now = self.client_now()
        return client_now.year

    def get_default_client_timezone(self):
        """
        gets the default client timezone.

        :rtype: tzinfo
        """

        return self._client_timezone

    def get_current_timezone(self, server):
        """
        gets the current timezone for server or client.

        :param bool server: if set to True, server timezone will be returned.
                            if set to False, client timezone will be returned.

        :rtype: tzinfo
        """

        if server is True:
            return self._server_timezone
        else:
            request = session_services.get_safe_current_request()
            timezone = None
            if request is not None:
                timezone = request.timezone

            return timezone or self._client_timezone

    def get_timezone(self, timezone):
        """
        gets the timezone based on given timezone name.

        :param str timezone: timezone name.

        :rtype: tzinfo
        """

        return pytz.timezone(timezone)

    def get_timezone_name(self, server):
        """
        gets the server or client timezone name.

        :param bool server: if set to True, server timezone name will be returned.
                            if set to False, client timezone name will be returned.

        :rtype: str
        """

        return self.get_current_timezone(server).zone

    def as_timezone(self, value, server):
        """
        gets the result of `astimezone` on the given value.

        :param datetime value: value to get normalized.

        :param bool server: if set to True, server timezone name will be used.
                            if set to False, client timezone name will be used.

        :rtype: datetime
        """

        timezone = self.get_current_timezone(server=server)
        return value.astimezone(timezone)

    def normalize(self, value, server):
        """
        normalizes input value using server or client current timezone and returns it.

        :param datetime value: value to get normalized.
        :param bool server: specifies that server or client timezone must used.

        :rtype: datetime
        """

        value = self.as_timezone(value, server=server)
        return self.get_current_timezone(server).normalize(value)

    def localize(self, value, server):
        """
        localizes input datetime with current timezone.

        input value should not have a timezone info.

        :param datetime value: value to be localized.
        :param bool server: specifies that server or client timezone must used.

        :rtype: datetime
        """

        return self.get_current_timezone(server).localize(value)

    def replace_timezone(self, value, server):
        """
        replaces given value's timezone with server or client timezone.

        it returns a new object.
        note that this method does not normalize the value, it just replaces the
        timezone and localizes the value.

        :param datetime value: value to replace its timezone.
        :param bool server: specifies that server or client timezone must used.

        :rtype: datetime
        """

        value = value.replace(tzinfo=None)
        return self.localize(value, server)

    def to_datetime_string(self, value, server):
        """
        gets the datetime string representation of input value.

        if the value has no timezone info, it adds the client or server
        timezone info based on `server` value.

        example: `2015-12-24T22:40:15+01:00`

        :param datetime value: input object to be converted.

        :param bool server: specifies that value must be normalized
                            to server or client timezone.

        :rtype: str
        """

        localized_value = self._add_timezone(value, server=server)
        localized_value = self.normalize(localized_value, server=server)
        return datetime_utils.to_datetime_string(localized_value)

    def to_date_string(self, value):
        """
        gets the date string representation of input value.

        example: `2015-12-24`

        :param datetime | date value: input object to be converted.

        :rtype: str
        """

        return datetime_utils.to_date_string(value)

    def to_time_string(self, value, server):
        """
        gets the time string representation of input value.

        if the value is a datetime and has no timezone info, it adds
        the client or server timezone info based on `server` value.

        example: `23:40:15`

        :param datetime | time value: input object to be converted.

        :param bool server: specifies that value must be normalized
                            to server or client timezone.

        :rtype: str
        """

        localized_value = value
        if isinstance(value, datetime):
            localized_value = self._add_timezone(value, server=server)
            localized_value = self.normalize(localized_value, server=server)

        return datetime_utils.to_time_string(localized_value)

    def to_datetime(self, value, server, replace_server=None):
        """
        converts the input value to it's equivalent python datetime.

        :param str value: string representation of datetime to be converted.

        :param bool server: specifies that value must be normalized
                            to server or client timezone.

        :param bool replace_server: specifies that it must replace the timezone
                                    of value with timezone of server before
                                    normalization. if set to False, it replaces
                                    it with client timezone.
                                    defaults to None and no replacement will be done.

        :rtype: datetime
        """

        converted_datetime = datetime_utils.to_datetime(value)
        if replace_server is not None:
            converted_datetime = self.replace_timezone(converted_datetime, replace_server)

        converted_datetime = self._add_timezone(converted_datetime, server=server)
        return self.normalize(converted_datetime, server=server)

    def to_date(self, value):
        """
        converts the input value to it's equivalent python date.

        :param str value: string representation of date to be converted.

        :rtype: date
        """

        return datetime_utils.to_date(value)

    def to_time(self, value):
        """
        converts the input value to it's equivalent python time.

        :param str value: string representation of time to be converted.

        :rtype: time
        """

        return datetime_utils.to_time(value)

    def timezone_exists(self, timezone_name):
        """
        gets a value indicating that a timezone with the given name exists.

        :param str timezone_name: timezone name to check for existence.

        :rtype: bool
        """

        return timezone_name in pytz.all_timezones_set

    def get_current_timestamp(self, date_sep='-', main_sep=' ',
                              time_sep=':', server=True, timezone=None):
        """
        gets the current timestamp with specified separators based on requested timezone.

        :param str date_sep: a separator to put between date elements.
        :param str main_sep: a separator to put between date and time part.
        :param str time_sep: a separator to put between time elements.

        :param bool server: if set to True, server timezone will be used.
                            if set to False, client timezone will be used.
                            defaults to True.

        :param str timezone: timezone name to get datetime based on it.
                             if provided, the value of `server` input
                             will be ignored. defaults to None.

        :rtype: str
        """

        if date_sep is None:
            date_sep = ''
        if main_sep is None:
            main_sep = ''
        if time_sep is None:
            time_sep = ''

        current = self.now(server=server, timezone=timezone)
        return '{year}{date_sep}{month}{date_sep}{day}{main_sep}' \
               '{hour}{time_sep}{minute}{time_sep}{second}'.format(
                year=str(current.year).zfill(4),
                month=str(current.month).zfill(2),
                day=str(current.day).zfill(2),
                hour=str(current.hour).zfill(2),
                minute=str(current.minute).zfill(2),
                second=str(current.second).zfill(2),
                date_sep=date_sep,
                main_sep=main_sep,
                time_sep=time_sep)



# -*- coding: utf-8 -*-
"""
serializer handlers tuple module.
"""

import pyrin.converters.serializer.services as serializer_services

from pyrin.converters.serializer.decorators import serializer
from pyrin.converters.serializer.handlers.base import SerializerBase
from pyrin.core.globals import ROW_RESULT


@serializer()
class TupleSerializer(SerializerBase):
    """
    tuple serializer class.
    """

    def _serialize(self, value, **options):
        """
        serializes the given value.

        returns serialized tuple on success or `NULL` object if serialization fails.

        :param tuple[object] value: tuple value to be serialized.

        :keyword ResultSchema result_schema: result schema instance to be
                                             used to create computed columns.
                                             defaults to None if not provided.

        :returns: serialized tuple of objects
        :rtype: tuple[dict]
        """

        if len(value) <= 0:
            return ()

        result = []
        for item in value:
            result.append(serializer_services.serialize(item, **options))

        return tuple(result)

    @property
    def accepted_type(self):
        """
        gets the accepted type for this serializer.

        which could serialize values from this type.

        :rtype: type[tuple]
        """

        return tuple

    def is_serializable(self, value, **options):
        """
        gets a value indicating that the given input is serializable.

        :param object value: value to be serialized.

        :rtype: bool
        """

        return super().is_serializable(value, **options) and not isinstance(value, ROW_RESULT)



# -*- coding: utf-8 -*-
"""
cli mixin handler module.
"""

from pyrin.cli.base import CLIHandlerBase
from pyrin.core.structs import CoreObject, DTO
from pyrin.utils.custom_print import print_warning
from pyrin.cli.exceptions import InvalidCLIHandlerTypeError, DuplicatedCLIHandlerError, \
    CLIHandlerNotFoundError


class CLIMixin(CoreObject):
    """
    cli mixin class.

    every class that needs to manage cli handlers must inherit from this.
    """

    _cli_handler_type = CLIHandlerBase

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CLIMixin.
        """

        super().__init__()

        # a dictionary containing cli handlers for different commands.
        # in the form of: {str handler_name: CLIHandlerBase handler}
        self._cli_handlers = DTO()

    def register_cli_handler(self, instance, **options):
        """
        registers a new cli handler or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a cli handler which is already registered.

        :param CLIHandlerBase instance: cli handler to be registered.
                                        it must be an instance of CLIHandlerBase.

        :keyword bool replace: specifies that if there is another registered
                               cli handler with the same name, replace it
                               with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidCLIHandlerTypeError: invalid cli handler type error.
        :raises DuplicatedCLIHandlerError: duplicated cli handler error.
        """

        if not isinstance(instance, self._cli_handler_type):
            raise InvalidCLIHandlerTypeError('Input parameter [{instance}] is '
                                             'not an instance of [{handler}].'
                                             .format(instance=instance,
                                                     handler=self._cli_handler_type))

        if instance.get_name() in self._cli_handlers:
            old_instance = self._cli_handlers.get(instance.get_name())
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedCLIHandlerError('There is another registered '
                                                'cli handler with name [{name}] '
                                                'but "replace" option is not set, so '
                                                'cli handler [{instance}] could not '
                                                'be registered.'
                                                .format(name=instance.get_name(),
                                                        instance=instance))

            print_warning('CLI handler [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._cli_handlers[instance.get_name()] = instance

    def _get_cli_handler(self, name):
        """
        gets a cli handler with the given name.

        if not available, it raises an error.

        :param str name: cli handler name to get its instance.

        :raises CLIHandlerNotFoundError: cli handler not found error.

        :rtype: CLIHandlerBase
        """

        if name not in self._cli_handlers:
            raise CLIHandlerNotFoundError('CLI handler [{name}] not found.'
                                          .format(name=name))

        return self._cli_handlers[name]

    def execute(self, handler_name, **options):
        """
        executes the handler with the given name with given inputs.

        :param str handler_name: handler name to be executed.

        :raises CLIHandlerNotFoundError: cli handler not found error.

        :rtype: int
        """

        handler = self._get_cli_handler(handler_name)
        return handler.execute(**options)



# -*- coding: utf-8 -*-
"""
security utils component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.security.utils import SecurityUtilsPackage
from pyrin.security.utils.manager import SecurityUtilsManager


@component(SecurityUtilsPackage.COMPONENT_NAME)
class SecurityUtilsComponent(Component, SecurityUtilsManager):
    """
    security utils component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
sentry integrations base module.
"""

from abc import abstractmethod

import pyrin.configuration.services as config_services

from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.logging.sentry.interface import AbstractSentryIntegration
from pyrin.logging.sentry.integrations.exceptions import SentryIntegrationNameRequiredError


class SentryIntegrationBase(AbstractSentryIntegration):
    """
    sentry integration base class.

    all application sentry integrations must be subclassed from this.
    """

    # name of the integration. each subclass must set this to a valid name.
    name = None

    def __init__(self, *args, **options):
        """
        initializes an instance of SentryIntegrationBase.

        :raises SentryIntegrationNameRequiredError: sentry integration name required error.
        """

        super().__init__()

        if self.name in (None, '') or self.name.isspace():
            raise SentryIntegrationNameRequiredError('Sentry integration name is required '
                                                     'for integration [{instance}].'
                                                     .format(instance=self))

        self._set_name(self.name)

    def configure(self, integrations, **options):
        """
        configures this integration.

        :param list[Integration] integrations: list of enabled integrations.
        """

        if self.is_enabled is not True:
            return

        config = self._configure(**options)
        integrations.append(config)

    @abstractmethod
    def _configure(self, **options):
        """
        configures this integration and gets the configured object.

        this method must be overridden in subclasses.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: Integration
        """

        raise CoreNotImplementedError()

    @property
    def is_enabled(self):
        """
        gets a value indicating that this integration is enabled in sentry config store.

        :rtype: bool
        """

        enabled = config_services.get('sentry', self.get_name(), 'enable')
        return enabled is True



# -*- coding: utf-8 -*-
"""
caching local containers package.
"""

from pyrin.packaging.base import Package


class CachingLocalContainersPackage(Package):
    """
    caching local containers package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
sequence package.
"""

from pyrin.packaging.base import Package


class SequencePackage(Package):
    """
    sequence package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'database.sequence.component'



# -*- coding: utf-8 -*-
"""
caching remote handlers redis module.
"""

import pickle

import redis

from pyrin.caching.decorators import cache
from pyrin.caching.remote.handlers.base import RemoteCacheBase
from pyrin.caching.remote.handlers.exceptions import HostnameOrUnixSocketRequiredError, \
    BothHostnameAndUnixSocketProvidedError


@cache()
class Redis(RemoteCacheBase):
    """
    redis class.
    """

    cache_name = 'redis'

    HIT_KEY = '__HIT_COUNT__'
    MISS_KEY = '__MISS_COUNT__'

    def __init__(self, *args, **options):
        """
        initializes an instance of Redis.

        :keyword int expire: default expire time of cached items in milliseconds.
                             if not provided, it will be get from `caching` config
                             store.

        :keyword bool consider_user: specifies that current user must also be
                                     included in cache key. if not provided, will
                                     be get from `caching` config store.

        :raises CacheNameIsRequiredError: cache name is required error.
        :raises InvalidCacheExpireTimeError: invalid cache expire time error.
        """

        super().__init__(*args, **options)

        self.client.set(self.HIT_KEY, 0)
        self.client.set(self.MISS_KEY, 0)

    def _create_client(self, *args, kwargs=None, **configs):
        """
        creates a client for connecting to redis server.

        :param object args: all positional arguments passed to `__init__` method.
        :param dict kwargs: all keyword arguments passed to `__init__` method.

        :keyword str host: the host name of redis server. it could be
                           ip, url or `localhost`.

        :keyword int port: port number of redis server.

        :keyword str unix_socket_path: path to unix socket file.

        :note host, port and unix_socket_path: both unix_socket_path and host, port
                                               could not be provided at the same time.

        :keyword int socket_connect_timeout: milliseconds to wait for a connection to
                                             the redis server. defaults to `forever`
                                             (uses the underlying default socket timeout,
                                             which can be very long).

        :keyword int socket_timeout: milliseconds to wait for send or recv calls on
                                     the socket connected to redis. defaults to
                                     `forever` (uses the underlying default socket
                                     timeout, which can be very long).

        :keyword bool retry_on_timeout: retry on tcp connection timeout for
                                        sending/receiving values.

        :keyword str encoding: controls data encoding.
        :keyword str encoding_errors: how to face encoding errors.
        :keyword int db: specify a database number.
        :keyword str username: tcp connection username.
        :keyword str password: tcp connection password.
        :keyword bool socket_keepalive: keep-alive tcp connection.
        :keyword dict socket_keepalive_options: tcp keep-alive options.
        :keyword bool decode_responses: decode responses.
        :keyword bool ssl: use ssl connection to server.
        :keyword str ssl_keyfile: path to ssl key file.
        :keyword str ssl_certfile: path to ssl cert file.
        :keyword str ssl_cert_reqs: require certificate in connection.
        :keyword str ssl_ca_certs: path to the CA certificate.
        :keyword bool ssl_check_hostname: ssl check hostname.
        :keyword int max_connections: max connections to redis server.
        :keyword bool single_connection_client: create a single connection client.
        :keyword int health_check_interval: health check interval in seconds.
        :keyword str client_name: redis client name.

        :raises BothHostnameAndUnixSocketProvidedError: both hostname and unix
                                                        socket provided error.

        :raises HostnameOrUnixSocketRequiredError: hostname or unix socket
                                                   required error.

        :rtype: redis.Redis
        """

        unix_socket_path = configs.get('unix_socket_path', None)
        host = configs.get('host', None)
        port = configs.get('port', None)

        if unix_socket_path is not None and (host is not None or port is not None):
            raise BothHostnameAndUnixSocketProvidedError('Both host and unix socket could '
                                                         'not be provided at the same time.')

        if unix_socket_path is None and (host is None or port is None):
            raise HostnameOrUnixSocketRequiredError('Host or unix socket must be provided '
                                                    'to connect to redis server.')

        socket_timeout = configs.get('socket_timeout')
        if socket_timeout is not None and socket_timeout > 0:
            socket_timeout = socket_timeout / 1000
            configs.update(socket_timeout=socket_timeout)

        socket_connect_timeout = configs.get('socket_connect_timeout')
        if socket_connect_timeout is not None and socket_connect_timeout > 0:
            socket_connect_timeout = socket_connect_timeout / 1000
            configs.update(socket_connect_timeout=socket_connect_timeout)

        client = redis.Redis(**configs)
        return client

    def _increase_hit(self):
        """
        increases hit count of this cache.
        """

        self.client.incr(self.HIT_KEY, 1)

    def _increase_miss(self):
        """
        increases miss count of this cache.
        """

        self.client.incr(self.MISS_KEY, 1)

    def _clear(self):
        """
        clears all items from cache.
        """

        self.client.flushdb(asynchronous=True)

    def _set(self, key, value, *args, **options):
        """
        sets a new value into cached items.

        :param object key: hashable key of the cache to be registered.
        :param object value: value to be cached.

        :keyword int expire: expire time for this item in milliseconds.
                             if not provided, it will be get from `expire` attribute.
        """

        expire = options.get('expire')
        if expire is None:
            expire = self.expire

        if expire == 0:
            expire = None

        self.client.set(key, value, px=expire)

    def _get(self, key, default=None, **options):
        """
        gets the value from cache.

        if key does not exist, it returns None or the specified default value.

        :param object key: hashable key to get its value from cache.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        result = self.client.get(key)
        if result is None:
            self._increase_miss()
            return default

        self._increase_hit()
        return result

    def _prepare_key(self, key):
        """
        prepares the key to be used in cache.

        :param object key: key to be cached.

        :returns: prepared key.
        :rtype: str
        """

        return str(key)

    def _prepare_for_set(self, value):
        """
        prepares the value that must be cached.

        :param object value: value to be cached.

        :returns: prepared value
        :rtype: bytes
        """

        return pickle.dumps(value)

    def _prepare_for_get(self, value):
        """
        prepares the value that must be returned from cache.

        :param bytes value: value to be returned.

        :returns: prepared value.
        """

        return pickle.loads(value)

    def increment(self, key, value):
        """
        increments the value of given key by provided value.

        if the key does not exist, it will be initialized by value.
        the cached and provided values must be integers.
        it returns the new value.

        :param object key: hashable key to increase its value in the cache.
        :param int value: the amount by which to increment the cached value.

        :rtype: int
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        return self.client.incr(hashed_key, value)

    def decrement(self, key, value):
        """
        decrements the value of given key by provided value.

        if the key does not exist, it will be initialized as 0.
        the cached and provided values must be integers.
        it returns the new value.

        :param object key: hashable key to decrease its value in the cache.
        :param int value: the amount by which to decrement the cached value.

        :rtype: int
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        return self.client.decr(hashed_key, value)

    def contains(self, key):
        """
        gets a value indicating that given key is existed in the cached items.

        :param object key: key to be checked for existence.

        :rtype: bool
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        result = self.client.exists(hashed_key)
        return result > 0

    def pop(self, key, default=None):
        """
        pops the given key from cache and returns its value.

        if key does not exist, it returns None or the specified default value.

        :param object key: key to get its value.
        :param object default: value to be returned if key is not present.

        :returns: object
        """

        result = self.get(key, default=default)
        self.remove(key)
        return result

    def remove(self, key, **options):
        """
        removes the given key from cache.

        it does nothing if the key is not present in the cache.

        :param object key: key to be removed.
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        self.client.delete(hashed_key)

    def touch(self, key):
        """
        extends the expire time of given key in the cache.

        :param object key: hashable key to extend its expire time.

        :returns: True if expire time updated, False if the key is not existed.
        :rtype: bool
        """

        hashed_key = hash(key)
        hashed_key = self._prepare_key(hashed_key)
        return self.client.touch(hashed_key)

    @property
    def stats(self):
        """
        get the statistic info about cached items.

        :returns: dict(datetime last_cleared_time: last cleared time,
                       bool persistent: persistent cache,
                       bool consider_user: consider user,
                       int expire: cached items expire time,
                       int hit: hit count,
                       int miss: miss count,
                       str hit_ratio: hit ratio,
                       dict memory_stats: memory stats)
        :rtype: dict
        """

        base_stats = super().stats
        stats = dict(memory_stats=self.client.memory_stats())
        base_stats.update(stats)
        return base_stats

    @property
    def hit_count(self):
        """
        gets the hit count for this cache.

        :rtype: int
        """

        return int(self.client.get(self.HIT_KEY) or 0)

    @property
    def miss_count(self):
        """
        gets the miss count for this cache.

        :rtype: int
        """

        return int(self.client.get(self.MISS_KEY) or 0)



# -*- coding: utf-8 -*-
"""
validator services module.
"""

from pyrin.application.services import get_component
from pyrin.validator import ValidatorPackage


def register_validator(instance, **options):
    """
    registers a new validator or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a validator which is already registered.

    :param AbstractValidatorBase instance: validator to be registered.
                                           it must be an instance of
                                           AbstractValidatorBase.

    :keyword bool replace: specifies that if there is another registered
                           validator with the same domain and name, replace
                           it with the new one, otherwise raise an error.
                           defaults to False.

    :raises InvalidValidatorTypeError: invalid validator type error.
    :raises DuplicatedValidatorError: duplicated validator error.
    """

    get_component(ValidatorPackage.COMPONENT_NAME).register_validator(instance, **options)


def get_domain_validators(domain):
    """
    gets all registered validators for given domain.

    it raises an error if domain does not exist.

    :param type[BaseEntity] | str domain: the domain to get its validators.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :raises ValidatorDomainNotFoundError: validator domain not found error.

    :rtype: dict[type[BaseEntity] | str, AbstractValidatorBase]
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).get_domain_validators(domain)


def get_validator(domain, name):
    """
    gets the registered validator for given domain and name.

    it returns None if no validator found for given name.

    :param type[BaseEntity] | str domain: the domain to get validator from.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :param str name: validator name to get.

    :raises ValidatorDomainNotFoundError: validator domain not found error.

    :rtype: AbstractValidatorBase
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).get_validator(domain, name)


def try_get_validator(domain, name):
    """
    gets the registered validator for given domain and name.

    it returns None if no validator found for given name or if domain does not exist.

    :param type[BaseEntity] | str domain: the domain to get validator from.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :param str name: validator name to get.

    :rtype: AbstractValidatorBase
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).try_get_validator(domain, name)


def validate_field(domain, name, value, **options):
    """
    validates the given value with given validator.

    it returns the same value or fixed one.

    :param type[BaseEntity] | str domain: the domain to validate the value for.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :param str name: validator name to be used for validation.
    :param object | list[object] value: value to be validated.

    :keyword bool force: specifies that if there is no validator
                         with specified domain and name, it should
                         raise an error. defaults to False if not provided.

    :keyword bool for_update: specifies that this field is being
                              validated for update operation.
                              defaults to False if not provided.

    :keyword bool nullable: determines that provided value could be None.
    :keyword bool is_list: specifies that the value must be a list of items.

    :keyword bool null_items: specifies that list items could be None.
                              it is only used if `is_list=True` is provided.

    :keyword bool allow_single: specifies that list validator should also
                                accept single, non list values.
                                it is only used if `is_list=True` is provided.
                                defaults to False if not provided.

    :keyword bool allow_empty_list: specifies that list validators should also
                                    accept empty lists.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

    :keyword bool inclusive_minimum: determines that values equal to
                                     accepted minimum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool inclusive_maximum: determines that values equal to
                                     accepted maximum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool allow_blank: determines that empty strings should be
                               considered valid. this argument will only
                               be considered in string validators.

    :keyword bool allow_whitespace: determines that whitespace strings should be
                                    considered valid. this argument will only
                                    be considered in string validators.

    :raises ValidatorDomainNotFoundError: validator domain not found error.
    :raises ValidatorNotFoundError: validator not found error.
    :raises ValidationError: validation error.

    :returns: same value or fixed one.
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).validate_field(domain, name,
                                                                         value, **options)


def validate_dict(domain, data, **options):
    """
    validates available values of given dict.

    it uses the correct validator for each value based on its key name.

    :param type[BaseEntity] | str domain: the domain to validate the values for.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :param dict data: dictionary to validate its values.

    :keyword bool lazy: specifies that all values must be validated first and
                        then a cumulative error must be raised containing a dict
                        of all keys and their corresponding error messages.
                        defaults to True if not provided.

    :keyword bool for_update: specifies that only fields that are present in data
                              must be validated. even if there are validators for
                              other fields. it is useful for changing the validation
                              behavior on insert or update operations. defaults to
                              False if not provided and all validators will be used.

    :keyword bool nullable: determines that provided values could be None.
    :keyword bool is_list: specifies that the value must be a list of items.

    :keyword bool null_items: specifies that list items could be None.
                              it is only used if `is_list=True` is provided.

    :keyword bool allow_single: specifies that list validator should also
                                accept single, non list values.
                                it is only used if `is_list=True` is provided.
                                defaults to False if not provided.

    :keyword bool allow_empty_list: specifies that list validators should also
                                    accept empty lists.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

    :keyword bool inclusive_minimum: determines that values equal to
                                     accepted minimum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool inclusive_maximum: determines that values equal to
                                     accepted maximum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool allow_blank: determines that empty strings should be
                               considered valid. this argument will only
                               be considered in string validators.

    :keyword bool allow_whitespace: determines that whitespace strings should be
                                    considered valid. this argument will only
                                    be considered in string validators.

    :keyword BaseEntity entity: an entity instance that the provided data
                                is the result dict of it.
                                it will be used to populate fixed values
                                in the entity.

    :raises InvalidDataForValidationError: invalid data for validation error.
    :raises ValidatorDomainNotFoundError: validator domain not found error.
    :raises ValidatorNotFoundError: validator not found error.
    :raises ValidationError: validation error.
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).validate_dict(domain, data,
                                                                        **options)


def validate_entity(entity, **options):
    """
    validates available values of given entity.

    it uses the correct validator for each value based on its field name.

    :param BaseEntity entity: entity to validate its values.

    :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                  which has `allow_read=False` or its name
                                                  starts with underscore `_`, should not
                                                  be included in validation. defaults to
                                                  `SECURE_TRUE` if not provided.

    :keyword bool lazy: specifies that all fields must be validated first and
                        then a cumulative error must be raised containing a dict
                        of all field names and their corresponding error messages.
                        defaults to True if not provided.

    :keyword bool nullable: determines that provided values could be None.
    :keyword bool is_list: specifies that the value must be a list of items.

    :keyword bool null_items: specifies that list items could be None.
                              it is only used if `is_list=True` is provided.

    :keyword bool allow_single: specifies that list validator should also
                                accept single, non list values.
                                it is only used if `is_list=True` is provided.
                                defaults to False if not provided.

    :keyword bool allow_empty_list: specifies that list validators should also
                                    accept empty lists.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

    :keyword bool inclusive_minimum: determines that values equal to
                                     accepted minimum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool inclusive_maximum: determines that values equal to
                                     accepted maximum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool allow_blank: determines that empty strings should be
                               considered valid. this argument will only
                               be considered in string validators.

    :keyword bool allow_whitespace: determines that whitespace strings should be
                                    considered valid. this argument will only
                                    be considered in string validators.

    :raises InvalidEntityForValidationError: invalid entity for validation error.
    :raises ValidatorDomainNotFoundError: validator domain not found error.
    :raises ValidatorNotFoundError: validator not found error.
    :raises ValidationError: validation error.
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).validate_entity(entity, **options)


def is_valid_field(domain, name, value, **options):
    """
    gets a value indicating that given field is valid.

    :param type[BaseEntity] | str domain: the domain to validate the value for.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :param str name: validator name to be used for validation.
    :param object | list[object] value: value to be validated.

    :keyword bool force: specifies that if there is no validator
                         with specified domain and name, it should
                         raise an error. defaults to False if not provided.

    :keyword bool for_update: specifies that this field is being
                              validated for update operation.
                              defaults to False if not provided.

    :keyword bool nullable: determines that provided value could be None.
    :keyword bool is_list: specifies that the value must be a list of items.

    :keyword bool null_items: specifies that list items could be None.
                              it is only used if `is_list=True` is provided.

    :keyword bool allow_single: specifies that list validator should also
                                accept single, non list values.
                                it is only used if `is_list=True` is provided.
                                defaults to False if not provided.

    :keyword bool allow_empty_list: specifies that list validators should also
                                    accept empty lists.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

    :keyword bool inclusive_minimum: determines that values equal to
                                     accepted minimum should be considered valid.
                                     this argument will only be considered in min
                                     and range validators.

    :keyword bool inclusive_maximum: determines that values equal to
                                     accepted maximum should be considered valid.
                                     this argument will only be considered in max
                                     and range validators.

    :keyword bool allow_blank: determines that empty strings should be
                               considered valid. this argument will only
                               be considered in string validators.

    :keyword bool allow_whitespace: determines that whitespace strings should be
                                    considered valid. this argument will only
                                    be considered in string validators.

    :raises ValidatorDomainNotFoundError: validator domain not found error.
    :raises ValidatorNotFoundError: validator not found error.

    :rtype: bool
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).is_valid_field(domain, name,
                                                                         value, **options)


def is_valid_dict(domain, data, **options):
    """
    gets a value indicating that given dict has valid values.

    it uses the correct validator for each value based on its key name.

    :param type[BaseEntity] | str domain: the domain to validate the values for.
                                          it could be a type of a BaseEntity
                                          subclass or a string name.

    :param dict data: dictionary to validate its values.

    :keyword bool for_update: specifies that only fields that are present in data
                              must be validated. even if there are validators for
                              other fields. it is useful for changing the validation
                              behavior on insert or update operations. defaults to
                              False if not provided and all validators will be used.

    :keyword bool nullable: determines that provided values could be None.
    :keyword bool is_list: specifies that the value must be a list of items.

    :keyword bool null_items: specifies that list items could be None.
                              it is only used if `is_list=True` is provided.

    :keyword bool allow_single: specifies that list validator should also
                                accept single, non list values.
                                it is only used if `is_list=True` is provided.
                                defaults to False if not provided.

    :keyword bool allow_empty_list: specifies that list validators should also
                                    accept empty lists.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

    :keyword bool inclusive_minimum: determines that values equal to
                                     accepted minimum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool inclusive_maximum: determines that values equal to
                                     accepted maximum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool allow_blank: determines that empty strings should be
                               considered valid. this argument will only
                               be considered in string validators.

    :keyword bool allow_whitespace: determines that whitespace strings should be
                                    considered valid. this argument will only
                                    be considered in string validators.

    :raises InvalidDataForValidationError: invalid data for validation error.
    :raises ValidatorDomainNotFoundError: validator domain not found error.
    :raises ValidatorNotFoundError: validator not found error.

    :rtype: bool
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).is_valid_dict(domain, data,
                                                                        **options)


def is_valid_entity(entity, **options):
    """
    gets a value indicating that given entity has valid values.

    it uses the correct validator for each value based on its field name.

    :param BaseEntity entity: entity to validate its values.

    :keyword SECURE_TRUE | SECURE_FALSE readable: specifies that any column or attribute
                                                  which has `allow_read=False` or its name
                                                  starts with underscore `_`, should not
                                                  be included in validation. defaults to
                                                  `SECURE_TRUE` if not provided.

    :keyword bool nullable: determines that provided values could be None.
    :keyword bool is_list: specifies that the value must be a list of items.

    :keyword bool null_items: specifies that list items could be None.
                              it is only used if `is_list=True` is provided.

    :keyword bool allow_single: specifies that list validator should also
                                accept single, non list values.
                                it is only used if `is_list=True` is provided.
                                defaults to False if not provided.

    :keyword bool allow_empty_list: specifies that list validators should also
                                    accept empty lists.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

    :keyword bool inclusive_minimum: determines that values equal to
                                     accepted minimum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool inclusive_maximum: determines that values equal to
                                     accepted maximum should be considered valid.
                                     this argument will only be considered in min,
                                     max and range validators.

    :keyword bool allow_blank: determines that empty strings should be
                               considered valid. this argument will only
                               be considered in string validators.

    :keyword bool allow_whitespace: determines that whitespace strings should be
                                    considered valid. this argument will only
                                    be considered in string validators.

    :raises InvalidEntityForValidationError: invalid entity for validation error.
    :raises ValidatorDomainNotFoundError: validator domain not found error.
    :raises ValidatorNotFoundError: validator not found error.

    :rtype: bool
    """

    return get_component(ValidatorPackage.COMPONENT_NAME).is_valid_entity(entity, **options)



# -*- coding: utf-8 -*-
"""
configuration store module.
"""

import os

from copy import deepcopy

import pyrin.converters.deserializer.services as deserializer_services
import pyrin.utils.configuration as config_utils

from pyrin.utils.exceptions import ConfigurationFileNotFoundError as UtilsFileNotFoundError
from pyrin.core.structs import CoreObject, DTO
from pyrin.utils.dictionary import change_key_case
from pyrin.configuration.exceptions import ConfigurationFileNotFoundError, \
    ConfigurationStoreKeyNotFoundError, ConfigurationStoreSectionNotFoundError, \
    ConfigurationStoreDuplicateKeyError, ConfigurationEnvironmentVariableNotFoundError, \
    InvalidConfigurationEnvironmentVariableValueError, \
    ConfigurationStoreAttributeNotFoundException, InvalidEnvironmentVariableKeyError, \
    InvalidPlaceHolderForEnvironmentVariableError


class ConfigStore(CoreObject):
    """
    config store class.

    all configurations will be stored with lowercase keys.
    """

    ACTIVE_SECTION_NAME = 'active'
    SELECTED_SECTION_NAME = 'selected'

    ENV_SEPARATOR = ':'
    # all values of different keys in a config store which start
    # with this, will be get from environment variables.
    ENV_PLACE_HOLDER = 'from_env' + ENV_SEPARATOR

    def __init__(self, name, config_file_path, **options):
        """
        initializes an instance of ConfigStore.

        :param str name: config store name.
        :param str config_file_path: full path of config file.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        super().__init__()

        self._configs = DTO()
        self._defaults = options.get('defaults', None)
        self._name = name
        self._config_file_path = config_file_path
        self._load(**options)

    def __str__(self):
        """
        gets the string representation of current store.

        :rtype: str
        """

        return '{base} [{store}]'.format(base=super().__str__(), store=self._name)

    def _load(self, **options):
        """
        loads configurations from config file path.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        try:
            self._configs = config_utils.load(self._config_file_path,
                                              deserializer_services.deserialize,
                                              **options)

        except UtilsFileNotFoundError as error:
            raise ConfigurationFileNotFoundError(error) from error

        self._sync_with_env(**options)

    def reload(self, **options):
        """
        reloads configuration from it's physical file path.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        defaults = options.get('defaults', None)
        if defaults is None:
            options.update(defaults=self._defaults)
        else:
            self._defaults = defaults

        self._configs.clear()
        self._load(**options)

    def get(self, section, key, **options):
        """
        gets the value of given key from provided section.

        :param str section: config section name.
        :param str key: config key to get it's value.

        :keyword object default: default value if key not present in config section.
                                 if not provided, error will be raised.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        section_data = self.get_section(section, **options)
        if key in section_data:
            return section_data[key]

        if 'default' not in options:
            raise ConfigurationStoreKeyNotFoundError('Key [{key}] not found in section '
                                                     '[{section}] from config store [{name}].'
                                                     .format(key=key,
                                                             section=section,
                                                             name=self._name))

        return options.get('default')

    def get_active(self, key, **options):
        """
        gets the value of given key from active section of this config store.

        if this store does not have an active section, it raises an error.

        :param str key: config key to get it's value.

        :keyword object default: default value if key not present in config section.
                                 if not provided, error will be raised.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        active_section = self.get_active_section_name()
        return self.get(active_section, key, **options)

    def get_section_names(self, **options):
        """
        gets all available section names of config store.

        :rtype: list[str]
        """

        return list(self._configs.keys())

    def get_section(self, section, **options):
        """
        gets all key/values stored in given section.

        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: dict
        """

        if section not in self._configs:
            raise ConfigurationStoreSectionNotFoundError('Section [{section}] not found in '
                                                         'config store [{name}].'
                                                         .format(section=section,
                                                                 name=self._name))

        result = deepcopy(self._configs.get(section))
        return self._change_key_case(result, **options)

    def get_section_keys(self, section, **options):
        """
        gets all available keys in given section.

        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: list[str]
        """

        return list(self.get_section(section, **options).keys())

    def get_all(self, **options):
        """
        gets all available key/values from different sections of this config store.

        in a flat dict, eliminating the sections.
        note that if there are same key names with different values
        in different sections, it raises an error to prevent overwriting
        values. also note that if this config store contains `active` section,
        then the result of `get_active_section` method would be returned.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreDuplicateKeyError: configuration store duplicate key error.

        :rtype: dict
        """

        # trying to get active config if available.
        active_config = None
        try:
            active_config = self.get_active_section(**options)
        except ConfigurationStoreAttributeNotFoundException:
            pass

        if active_config is not None:
            return active_config

        flat_dict = DTO()
        for section_data in self._get_sections(**options):
            if isinstance(section_data, dict):
                for key, value in section_data.items():
                    if key in flat_dict and value != flat_dict.get(key):
                        raise ConfigurationStoreDuplicateKeyError('Key [{key}] is available '
                                                                  'in multiple sections of '
                                                                  'config store [{name}] '
                                                                  'with different values.'
                                                                  .format(key=key,
                                                                          name=self._name))
                    flat_dict[key] = value

        return flat_dict

    def get_file_path(self, **options):
        """
        gets config file path of this config store.

        :rtype: str
        """

        return self._config_file_path

    def _change_key_case(self, value, **options):
        """
        returns a copy of input dict with all it's keys
        and nested keys cases modified using given converter.
        if converter is not provided, it returns the inputted dict.

        :param dict value: dict to change it's keys cases.

        :keyword callable converter: a callable to use as case converter.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :rtype: dict
        """

        converter = options.get('converter', None)
        if converter is not None:
            return change_key_case(value, converter)

        return value

    def get_active_section(self, **options):
        """
        gets the active section available in related config file.

        this method gets the section that it's name is under [active]
        section, for example:

        [active]
        selected: production

        [production]
        id: 123
        name: prod

        [development]
        id: 233
        name: dev

        this will return all key/values available under [production].
        if the config store has not an [active] section, this method
        raises an error.

        :keyword callable converter: a callable to use as case converter.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: dict
        """

        active_section = self.get_active_section_name()
        return self.get_section(active_section, **options)

    def _get_sections(self, **options):
        """
        gets a list of dictionaries containing all key/values
        for every section of this config store.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :rtype: list[dict]
        """

        return [self.get_section(section, **options)
                for section in self.get_section_names(**options)]

    def _get_from_env(self, key, **options):
        """
        gets the value of given key from environment variable if available.

        otherwise may raise an exception or ignore it depending on `silent` keyword.

        :param str key: key to get it's value from environment variable.

        :keyword bool silent: indicates that if an environment variable for the
                              given key not found, ignore it and return None, otherwise
                              raise an error. defaults to True if not provided.

        :raises ConfigurationEnvironmentVariableNotFoundError: configuration environment
                                                               variable not found error.

        :raises InvalidConfigurationEnvironmentVariableValueError: invalid configuration
                                                                   environment variable
                                                                   value error.

        :rtype: str
        """

        value = os.environ.get(key)
        silent = options.get('silent', True)
        if value is None and silent is not True:
            raise ConfigurationEnvironmentVariableNotFoundError('Configuration environment '
                                                                'variable [{key}] not found.'
                                                                .format(key=key))

        if value is not None and len(value.strip()) == 0 and silent is not True:
            raise InvalidConfigurationEnvironmentVariableValueError('Configuration environment '
                                                                    'variable [{key}] has '
                                                                    'an invalid value.'
                                                                    .format(key=key))

        return value

    def _sync_with_env(self, **options):
        """
        synchronizes all keys with None value of this config store.

        with environment variables if available. it searches for environment
        variables with name `STORE_SECTION_KEY`. for example `API_GENERAL_NAMESPACE`.
        it also updates all keys that their values start with `ENV_PLACE_HOLDER`
        from environment variables.

        :keyword bool silent: indicates that if an environment variable for the
                              config key not found, ignore it and return None, otherwise
                              raise an error. defaults to True.
                              note that this option is only for keys that have None value.
                              but those keys that have `ENV_PLACE_HOLDER` will always cause
                              an error if environment variable not found for them.

        :raises ConfigurationEnvironmentVariableNotFoundError: configuration environment
                                                               variable not found error.

        :raises InvalidConfigurationEnvironmentVariableValueError: invalid configuration
                                                                   environment variable
                                                                   value error.

        :raises InvalidPlaceHolderForEnvironmentVariableError: invalid place holder for
                                                               environment variable error.

        :raises InvalidEnvironmentVariableKeyError: invalid environment variable key error.
        """

        for section_name in self.get_section_names():
            section = self.get_section(section_name)
            for key, value in section.items():
                if self._should_try_env(value) is True:
                    env_value = None
                    if value is None:
                        key_name = self._get_env_key_name(self._name, section_name, key)
                        env_value = self._get_from_env(key_name, **options)
                    else:
                        env_key = self._extract_env_key(section_name, key, value)
                        env_value = self._get_from_env(env_key, silent=False)
                    converted_value = deserializer_services.deserialize(env_value)
                    self._configs[section_name][key] = converted_value

    def _get_env_key_name(self, store, section, key):
        """
        gets key name for given inputs to get it from environment variables.

        it generates key name in this format: `STORE_SECTION_KEY`.
        for example: `API_GENERAL_NAMESPACE`.

        :param str store: config store name.
        :param str section: config store section name.
        :param str key: section key name.

        :rtype: str
        """

        return '{store}_{section}_{key}'.format(store=store.upper(),
                                                section=section.upper(),
                                                key=key.upper())

    def _extract_env_key(self, section, key, value):
        """
        extracts environment variable key name from given value.

        :param str section: config section name.
        :param str key: key to get it's value from environment variable.
        :param str value: value to get environment variable key name from it.

        :raises InvalidPlaceHolderForEnvironmentVariableError: invalid place holder for
                                                               environment variable error.

        :raises InvalidEnvironmentVariableKeyError: invalid environment variable key error.

        :rtype: str
        """

        if self.ENV_SEPARATOR not in value:
            raise InvalidPlaceHolderForEnvironmentVariableError('Environment variable place '
                                                                'holder must contain a [{sep}] '
                                                                'character. but value [{value}] '
                                                                'for key [{key}] of section '
                                                                '[{section}] in config store '
                                                                '[{store}] does not contain any '
                                                                'of it.'
                                                                .format(sep=self.ENV_SEPARATOR,
                                                                        value=value, key=key,
                                                                        section=section,
                                                                        store=self))
        separator_index = value.index(self.ENV_SEPARATOR)
        env_key = value[separator_index + 1:]
        if env_key in (None, ''):
            raise InvalidEnvironmentVariableKeyError('Provided environment variable name is '
                                                     'invalid for config key [{key}] of section '
                                                     '[{section}] in config store [{store}].'
                                                     .format(key=key, section=section,
                                                             store=self))
        return env_key

    def _should_try_env(self, value):
        """
        gets a value indicating that given key could be get from environment variables.

        if the value is None or it starts with 'ENV_PLACE_HOLDER' this method returns true.

        :param object value: current value of given key in config file.

        :rtype: bool
        """

        return value is None or self._is_env_required(value)

    def _is_env_required(self, value):
        """
        gets a value indicating that given key must be get from environment variables.

        if the value starts with 'ENV_PLACE_HOLDER' this method returns true.

        :param object value: current value of given key in config file.

        :rtype: bool
        """

        return isinstance(value, str) and value.startswith(self.ENV_PLACE_HOLDER)

    def get_active_section_name(self):
        """
        gets the active section name of this config store if available.

        if the store does not have an active section, it raises an error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: str
        """

        return self.get(self.ACTIVE_SECTION_NAME, self.SELECTED_SECTION_NAME)

    def get_all_sections(self, **options):
        """
        gets all sections and their keys of current config store.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :rtype: dict
        """

        return deepcopy(self._configs)



# -*- coding: utf-8 -*-
"""
configuration manager module.
"""

import os

import pyrin.application.services as application_services
import pyrin.utils.path as path_utils
import pyrin.utils.file as file_utils

from pyrin.configuration import ConfigurationPackage
from pyrin.configuration.store import ConfigStore
from pyrin.core.structs import Context, Manager
from pyrin.configuration.exceptions import ConfigurationStoreExistedError, \
    ConfigurationStoreNotFoundError, ConfigurationFileNotFoundError, ConfigurationFileExistedError


class ConfigurationManager(Manager):
    """
    configuration manager class.
    """

    # config extension to be used for config file names.
    CONFIG_EXTENSION = '.ini'

    # class type to be used as config store.
    config_store_class = ConfigStore
    package_class = ConfigurationPackage

    def __init__(self, **options):
        """
        initializes an instance of ConfigurationManager.
        """

        super().__init__()

        self._config_stores = Context()
        self._settings_path = application_services.get_settings_path()
        self._default_settings_path = application_services.get_default_settings_path()

    def _add_config_store(self, name, file_path, **options):
        """
        adds a new config store for given file with the specified name.

        :param str name: config store name.
        :param str file_path: config file full path.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :keyword bool ignore_on_existed: specifies that it should not raise an
                                         error if a config store with given name
                                         has been already loaded.
                                         defaults to False if not provided.

        :raises ConfigurationStoreExistedError: configuration store existed error.
        """

        ignore = options.get('ignore_on_existed', False)
        if name in self._config_stores:
            if ignore is not True:
                raise ConfigurationStoreExistedError('Config store name [{name}] already '
                                                     'existed, config store names must be unique.'
                                                     .format(name=name))
        else:
            self._config_stores[name] = self.config_store_class(name, file_path, **options)

    def _is_config_file(self, file_name):
        """
        gets a value indicating that given file name belongs to a config file.

        :param str file_name: file name.

        :rtype: bool
        """

        return file_name.endswith(self.CONFIG_EXTENSION)

    def load_configuration(self, name, **options):
        """
        loads the given configuration if relevant file is available in settings path.

        it creates the config file in application settings path if
        config file is only available in pyrin default settings.

        :param str name: configuration name.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :keyword bool silent: specifies that if a related configuration file
                              for the given name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :keyword bool ignore_on_existed: specifies that it should not raise an
                                         error if a config store with given name
                                         has been already loaded.
                                         defaults to False if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        :raises ConfigurationStoreExistedError: configuration store existed error.
        """

        file_path = self._get_relevant_file_path(name, **options)
        if file_path is not None:
            self._add_config_store(name, file_path, **options)

            options.update(ignore_on_existed=True, silent=True)
            self.create_config_file(name, **options)

    def load_configurations(self, *names, **options):
        """
        loads the given configurations if relevant files is available in settings path.

        it creates the config files in application settings path if
        config files are only available in pyrin default settings.

        :param str names: configuration names as arguments.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :keyword bool silent: specifies that if a related configuration file
                              for any of the given names not found, ignore it.
                              otherwise raise an error. defaults to False.

        :keyword bool ignore_on_existed: specifies that it should not raise an
                                         error if a config store with given name
                                         has been already loaded.
                                         defaults to False if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        :raises ConfigurationStoreExistedError: configuration store existed error.
        """

        for single_name in names:
            self.load_configuration(single_name, **options)

    def _get_relevant_file_path(self, name, **options):
        """
        gets the relevant file path to specified store name in settings folder.

        it returns None if the file is not available and `silent=True` is given.

        :param str name: config store name.

        :keyword bool only_default: specifies that it should return the file path only
                                    if it's available in default settings directory.
                                    otherwise it returns the file path from application
                                    settings if available. if not, returns it from default
                                    settings directory. defaults to False if not provided.

        :keyword bool silent: specifies that if a related configuration file
                              for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        only_default = options.get('only_default', False)
        paths = (self._settings_path, self._default_settings_path)
        if only_default is True:
            paths = (self._default_settings_path, )

        config_name = self._get_file_name(name)
        config_path = path_utils.get_first_available_file(*paths, file_name=config_name)

        silent = options.get('silent', False)
        if config_path is None and silent is not True:
            raise ConfigurationFileNotFoundError('Config store name [{name}] does not '
                                                 'have any related configuration '
                                                 'file in application settings.'
                                                 .format(name=name))
        return config_path

    def _get_file_name(self, store):
        """
        gets the generic file name for given store name.

        note that the purpose of this method is different from `get_file_name()`
        method, this method makes a generic file name without any assertion
        for the file to be really existed. if you want to assert the existence
        of the file, you should use `get_file_name()` method.

        :param str store: store name to make file name for it.

        :rtype: str
        """

        return '{store}{extension}'.format(store=store, extension=self.CONFIG_EXTENSION)

    def reload(self, store_name, **options):
        """
        reloads the configuration store from it's relevant file.

        :param str store_name: config store name to be reloaded.

        :keyword dict defaults: a dict containing values
                                needed for interpolation.
                                defaults to None if not provided.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.
        """

        self._get_config_store(store_name).reload(**options)

    def get_default_file_path(self, store_name, **options):
        """
        gets the default configuration file path for given config store.

        it gets the file path in `pyrin.settings.default` directory.
        if the file is not available, it could return None or raise an error.

        :param str store_name: config store name to get its file path.

        :keyword bool silent: specifies that if a related default configuration
                              file for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        options.update(only_default=True)
        return self._get_relevant_file_path(store_name, **options)

    def get_file_path(self, store_name, **options):
        """
        gets the configuration file path for given config store.

        :param str store_name: config store name to get its file path.

        :keyword bool silent: specifies that if a related configuration file
                              for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        if store_name in self._config_stores.keys():
            return self._get_config_store(store_name).get_file_path(**options)

        return self._get_relevant_file_path(store_name, **options)

    def get_file_name(self, store_name, **options):
        """
        gets the configuration file name for given config store.

        for example: `database.ini`

        :param str store_name: config store name to get its file name.

        :keyword bool silent: specifies that if a related configuration file
                              for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        file_path = self.get_file_path(store_name, **options)
        if file_path is not None:
            return os.path.basename(file_path)

        return None

    def create_config_files(self, *stores, **options):
        """
        creates the config files for given store names in application settings path.

        it creates the files based on files available in pyrin default settings.

        :param str stores: store names to create config file for them.

        :keyword bool replace_existing: specifies that it should replace file if a config
                                        file is already available for given store name.
                                        this argument takes precedence over
                                        `ignore_on_existed` argument.
                                        defaults to False if not provided.

        :keyword bool ignore_on_existed: specifies that if a config file for the
                                         store name is already existed, ignore
                                         it, otherwise raise an error.
                                         defaults to False if not provided.

        :keyword bool silent: specifies that if a related default configuration
                              file for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :keyword dict[str, str] data: keyword with values to be replaced in config file.
                                      any keyword that does not exist in config file, will
                                      be ignored. defaults to None if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        :raises ConfigurationFileExistedError: configuration file existed error.
        """

        for store_name in stores:
            self.create_config_file(store_name, **options)

    def create_config_file(self, store, **options):
        """
        creates the config file for given store name in application settings path.

        it creates the file based on file available in pyrin default settings.

        :param str store: store name to create config file for it.

        :keyword bool replace_existing: specifies that it should replace file if a config
                                        file is already available for given store name.
                                        this argument takes precedence over
                                        `ignore_on_existed` argument.
                                        defaults to False if not provided.

        :keyword bool ignore_on_existed: specifies that if a config file for the
                                         store name is already existed, ignore
                                         it, otherwise raise an error.
                                         defaults to False if not provided.

        :keyword bool silent: specifies that if a related default configuration
                              file for the given store name not found, ignore it.
                              otherwise raise an error. defaults to False.

        :keyword dict[str, str] data: keyword with values to be replaced in config file.
                                      any keyword that does not exist in config file, will
                                      be ignored. defaults to None if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        :raises ConfigurationFileExistedError: configuration file existed error.
        """

        source_path = self.get_default_file_path(store, **options)
        if source_path is None:
            return

        config_name = self._get_file_name(store)
        config_path = os.path.abspath(os.path.join(self._settings_path, config_name))
        replace_existing = options.get('replace_existing', False)
        ignore_on_existed = options.get('ignore_on_existed', False)
        data = options.get('data', None)

        existed = os.path.exists(config_path)
        if existed and replace_existing is not True:
            if ignore_on_existed is not True:
                raise ConfigurationFileExistedError('Configuration file for store [{store}] '
                                                    'is already existed in [{file_path}].'
                                                    .format(store=store, file_path=config_path))
        else:
            if existed is True:
                path_utils.remove_file(config_path)

            if os.path.exists(self._settings_path) is False:
                path_utils.create_directory(self._settings_path)

            path_utils.copy_file(source_path, config_path)
            self._replace_config_values(config_path, data)

    def get(self, store_name, section, key, **options):
        """
        gets the value of specified key from provided section of given config store.

        :param str store_name: config store name.
        :param str section: config section name.
        :param str key: config key to get it's value.

        :keyword object default: default value if key not present in config section.
                                 if not provided, error will be raised.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        return self._get_config_store(store_name).get(section, key, **options)

    def get_active(self, store_name, key, **options):
        """
        gets the value of given key from active section of given config store.

        if this store does not have an active section, it raises an error.

        :param str store_name: config store name.
        :param str key: config key to get it's value.

        :keyword object default: default value if key not present in config section.
                                 if not provided, error will be raised.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.
        """

        return self._get_config_store(store_name).get_active(key, **options)

    def get_section_names(self, store_name, **options):
        """
        gets all available section names of given config store.

        :param str store_name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :rtype: list[str]
        """

        return self._get_config_store(store_name).get_section_names(**options)

    def get_section(self, store_name, section, **options):
        """
        gets all key/values stored in given section of specified config store.

        :param str store_name: config store name.
        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_section(section, **options)

    def get_section_keys(self, store_name, section, **options):
        """
        gets all available keys in given section of specified config store.

        :param str store_name: config store name.
        :param str section: section name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :rtype: list[str]
        """

        return self._get_config_store(store_name).get_section_keys(section, **options)

    def get_all(self, store_name, **options):
        """
        gets all available key/values from different sections of given config store.

        in a flat dict, eliminating the sections.
        note that if there are same key names with different values
        in different sections, it raises an error to prevent overwriting
        values. also note that if given config store contains `active` section,
        then the result of `get_active_section` method would be returned.

        :param str store_name: config store name.

        :keyword callable converter: a callable to use as case converter for keys.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.
        :raises ConfigurationStoreDuplicateKeyError: configuration store duplicate key error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_all(**options)

    def get_active_section(self, store_name, **options):
        """
        gets the active section available in given config store.

        this method gets the section that it's name is under [active]
        section, for example:

        [active]
        selected: production

        [production]
        id: 123
        name: prod

        [development]
        id: 233
        name: dev

        this will return all key/values available under [production].
        if the config store has not an [active] section, this method
        raises an error.

        :param str store_name: config store name.

        :keyword callable converter: a callable to use as case converter.
                                     it should be a callable with a signature
                                     similar to below example:
                                     case_converter(input_dict).

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store section
                                                        not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_active_section(**options)

    def get_active_section_name(self, store_name):
        """
        gets the active section name of given config store if available.

        if the store does not have an active section, it raises an error.

        :param str store_name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :raises ConfigurationStoreSectionNotFoundError: configuration store
                                                        section not found error.

        :raises ConfigurationStoreKeyNotFoundError: configuration store
                                                    key not found error.

        :rtype: str
        """

        return self._get_config_store(store_name).get_active_section_name()

    def _get_config_store(self, name):
        """
        gets the specified config store.

        :param str name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :rtype: ConfigStore
        """

        if name not in self._config_stores.keys():
            raise ConfigurationStoreNotFoundError('Config store [{name}] not found.'
                                                  .format(name=name))

        return self._config_stores[name]

    def get_all_sections(self, store_name, **options):
        """
        gets all sections and their keys of given config store.

        :param str store_name: config store name.

        :raises ConfigurationStoreNotFoundError: configuration store not found error.

        :rtype: dict
        """

        return self._get_config_store(store_name).get_all_sections(**options)

    def _replace_config_values(self, config_file, data):
        """
        replaces all config values available in given config file.

        with values provided in keyword arguments. the keys in kwargs, must
        be the same as the keys available in config file to be replaced.

        :param str config_file: absolute path to config file.

        :param dict[str, str] data: keyword with values to be replaced in config file.
                                    any keyword that does not exist in config file, will
                                    be ignored. defaults to None if not provided.

        :raises ConfigurationFileNotFoundError: configuration file not found error.
        """

        if not os.path.exists(config_file):
            raise ConfigurationFileNotFoundError('Config file [{file}] does not '
                                                 'exist in application settings.'
                                                 .format(file=config_file))

        result_dict = {}
        if data is not None and len(data) > 0:
            for name, value in data.items():
                regex = r'^{name}( )*[:=]{{1}}( )*(.*)$'.format(name=name)
                replace = '{name}: {value}'.format(name=name, value=value)
                result_dict[regex] = replace

            file_utils.replace_file_values_regex(config_file, result_dict)



# -*- coding: utf-8 -*-
"""
response package.
"""

from pyrin.packaging.base import Package


class ResponsePackage(Package):
    """
    response package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'processor.response.component'



# -*- coding: utf-8 -*-
"""
settings static module.
these settings are intended for values used inside some classes,
to be available before server startup.
"""

# default response status code.
DEFAULT_STATUS_CODE = 200

# default application encoding.
APPLICATION_ENCODING = 'utf-8'

# this value will be used to register default components that does not belong
# to any custom implementation. any other custom implementation that needs to be
# exposed through services, should provide it's own relevant key.
# example for COMPONENT_ID:
# (__name__, DEFAULT_COMPONENT_KEY) -> default
# (__name__, 4) -> custom
DEFAULT_COMPONENT_KEY = 0



# -*- coding: utf-8 -*-
"""
caching remote handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException


class CachingRemoteHandlersException(CoreException):
    """
    caching remote handlers exception.
    """
    pass


class BothHostnameAndUnixSocketProvidedError(CachingRemoteHandlersException):
    """
    both hostname and unix socket provided error.
    """
    pass


class HostnameOrUnixSocketRequiredError(CachingRemoteHandlersException):
    """
    hostname or unix socket required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
validator auto handler module.
"""

from pyrin.validator.interface import AbstractValidatorBase


class AutoValidator(AbstractValidatorBase):
    """
    auto validator class.
    """

    def __init__(self, domain, field, **options):
        """
        initializes an instance of AutoValidator.

        :param type[BaseEntity] domain: the domain in which this validator
                                        must be registered. it must be a
                                        type of a BaseEntity subclass.

        :param InstrumentedAttribute field: validator field. it must be a
                                            column. each validator will be registered
                                            with its field in corresponding domain.

        :keyword AbstractValidatorBase type_validator: the type validator instance.

        :keyword AbstractValidatorBase range_validator: the range validator instance.
                                                        it is either a range validator,
                                                        max validator, min validator or None.

        :keyword AbstractValidatorBase in_validator: the in validator instance.
                                                     it is either an in validator,
                                                     not in validator or None.

        :keyword AbstractValidatorBase custom_validator: a custom validator to extend
                                                         auto validator behavior.
        """

        super().__init__()

        self._name = field.key
        self._domain = domain
        self._type_validator = options.get('type_validator')
        self._range_validator = options.get('range_validator')
        self._in_validator = options.get('in_validator')
        self._custom_validator = options.get('custom_validator')

    def validate(self, value, **options):
        """
        validates the given value.

        it raises an error if validation fails.
        it returns the same or fixed value.

        :param object | list[object] value: value to be validated.

        :keyword bool for_update: specifies that this field is being
                                  validated for update operation.
                                  defaults to False if not provided.

        :raises ValidationError: validation error.

        :returns: object | list[object]
        """

        fixed_value = value
        if self._type_validator is not None:
            fixed_value = self._type_validator.validate(fixed_value, **options)

        if self._range_validator is not None:
            fixed_value = self._range_validator.validate(fixed_value, **options)

        if self._in_validator is not None:
            fixed_value = self._in_validator.validate(fixed_value, **options)

        if self._custom_validator is not None:
            fixed_value = self._custom_validator.validate(fixed_value, **options)

        return fixed_value

    @property
    def name(self):
        """
        gets the name of this validator.

        :rtype: str
        """

        return self._name

    @property
    def domain(self):
        """
        gets the domain of this validator.

        domain is the type of a BaseEntity subclass that
        this validator validates a value of it.

        :rtype: type[BaseEntity]
        """

        return self._domain



# -*- coding: utf-8 -*-
"""
sentry integrations exceptions module.
"""

from pyrin.core.exceptions import CoreException


class SentryIntegrationsException(CoreException):
    """
    sentry integrations exception.
    """
    pass


class SentryIntegrationNameRequiredError(SentryIntegrationsException):
    """
    sentry integration name required error.
    """
    pass



# -*- coding: utf-8 -*-
"""
database component module.
"""

from pyrin.application.decorators import component
from pyrin.database import DatabasePackage
from pyrin.database.manager import DatabaseManager
from pyrin.application.structs import Component


@component(DatabasePackage.COMPONENT_NAME)
class DatabaseComponent(Component, DatabaseManager):
    """
    database component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
deserializer decorators module.
"""

import pyrin.converters.deserializer.services as deserializer_services


def deserializer(*args, **kwargs):
    """
    decorator to register a deserializer.

    :param object args: deserializer class constructor arguments.
    :param object kwargs: deserializer class constructor keyword arguments.

    :keyword bool replace: specifies that if there is another registered
                           deserializer with the same name and accepted type,
                           replace it with the new one, otherwise raise
                           an error. defaults to False.

    :raises InvalidDeserializerTypeError: invalid deserializer type error.
    :raises DuplicatedDeserializerError: duplicated deserializer error.

    :returns: deserializer class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available deserializers.

        :param type cls: deserializer class.

        :returns: deserializer class.
        :rtype: type
        """

        instance = cls(*args, **kwargs)
        deserializer_services.register_deserializer(instance, **kwargs)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
celery cli handlers shell module.
"""

from pyrin.task_queues.celery.cli.decorators import celery_cli_handler
from pyrin.task_queues.celery.cli.enumerations import CeleryCLIHandlersEnum
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase
from pyrin.task_queues.celery.cli.handlers.params import IPythonParam, \
    BPythonParam, PythonParam, WithoutTasksParam, EventletParam, GeventParam


@celery_cli_handler()
class ShellCLIHandler(CeleryCLIHandlerBase):
    """
    shell cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of ShellCLIHandler.
        """

        super().__init__(CeleryCLIHandlersEnum.SHELL)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([IPythonParam(), BPythonParam(), PythonParam(),
                       WithoutTasksParam(), EventletParam(), GeventParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
orm sql operators base module.
"""

from sqlalchemy.sql import Select
from sqlalchemy.sql.elements import BindParameter
from sqlalchemy.sql.operators import ColumnOperators

import pyrin.utils.datetime as datetime_utils
import pyrin.utils.misc as misc_utils

from pyrin.utils.sqlalchemy import like_prefix, like_exact_prefix, \
    like_suffix, like_exact_suffix


class CoreColumnOperators(ColumnOperators):
    """
    core column operators class.

    this class provides some practical benefits
    to its subclasses for sql operations.
    """

    DEFAULT_ESCAPE_CHAR = '/'

    def _process_like_autoescape(self, value, escape, autoescape):
        """
        processes the value based on autoescape flag for like command.

        it may return an escaped string or the exact input string.

        :param str value: value to be escaped.

        :param str escape: a character to be used as escape character.
                           defaults to `DEFAULT_ESCAPE_CHAR` if not provided.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :rtype: str
        """

        if autoescape is True:
            if escape is None:
                escape = self.DEFAULT_ESCAPE_CHAR

            if escape not in ("%", "_"):
                value = value.replace(escape, escape + escape)

            value = value.replace("%", escape + "%").replace("_", escape + "_")

        return value

    def _process_like_prefix(self, value, **options):
        """
        processes the value that should be prefixed to

        value for `like` expression based on given options.

        :param str value: expression to be compared.

        :keyword bool exact_start: specifies that value should be emitted
                                   without any modifications at start of it.
                                   defaults to True if not provided.

        :keyword int start_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :note start_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                           if the provided value goes upper than this limit,
                           a `%` will be attached instead of it. this limit
                           is for security reason.

        :rtype: str
        """

        exact_start = options.get('exact_start', True)
        if exact_start is False:
            begin_wrapper = like_prefix
            inputs = (value, )
            begin_count = options.get('start_count', None)
            if begin_count is not None:
                begin_wrapper = like_exact_prefix
                inputs = (value, begin_count)

            value = begin_wrapper(*inputs)

        return value

    def _process_like_suffix(self, value, **options):
        """
        processes the value that should be suffixed to
        value for `like` expression based on given options.

        :param str value: expression to be compared.

        :keyword bool exact_end: specifies that value should be emitted
                                 without any modifications at end of it.
                                 defaults to True if not provided.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                         if the provided value goes upper than this limit,
                         a `%` will be attached instead of it. this limit
                         is for security reason.

        :rtype: str
        """

        exact_end = options.get('exact_end', True)
        if exact_end is False:
            end_wrapper = like_suffix
            inputs = (value,)
            end_count = options.get('end_count', None)
            if end_count is not None:
                end_wrapper = like_exact_suffix
                inputs = (value, end_count)

            value = end_wrapper(*inputs)

        return value

    def between_date(self, cleft, cright, symmetric=False, **options):
        """
        produces an `expression.between` clause against the parent
        object, given the lower and upper datetime range. this method is
        implemented to be able to handle datetime values more practical.

        :param object cleft: lower bound of datetime clause.
        :param object cright: upper bound of datetime clause.

        :param bool symmetric: specifies to emmit `between symmetric` to database.
                               note that not all databases support symmetric.
                               but `between symmetric` is equivalent to
                               `between least(a, b) and greatest(a, b)`.

        :keyword bool consider_begin_of_day: specifies that consider begin
                                             of day for lower datetime.
                                             defaults to True if not provided.

        :keyword bool consider_end_of_day: specifies that consider end
                                           of day for upper datetime.
                                           defaults to True if not provided.
        """

        consider_begin_of_day = options.get('consider_begin_of_day', True)
        consider_end_of_day = options.get('consider_end_of_day', True)

        if consider_begin_of_day is True:
            cleft = datetime_utils.begin_of_day(cleft)

        if consider_end_of_day is True:
            cright = datetime_utils.end_of_day(cright)

        # swapping values in case of user mistake.
        if type(cleft) is type(cright) and cleft > cright:
            cleft, cright = cright, cleft
            symmetric = False

        return self.between(cleft, cright, symmetric=symmetric)

    def istartswith(self, other, **options):
        """
        implements the `startswith` operator.

        produces an `ilike` expression that tests against a match for the
        start of a string value. for example:
        column like <other> || `%` or column like <other> || `_`
        this method provides a case-insensitive variant of `startswith()` method.

        :param str other: expression to be compared.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                         if the provided value goes upper than this limit,
                         a `%` will be attached instead of it. this limit
                         is for security reason.
        """

        escape = options.pop('escape', None)
        options.update(exact_end=False)
        other = self._process_like_autoescape(other, escape, options.get('autoescape', False))
        other = self._process_like_suffix(other, **options)
        return self.ilike(other, escape)

    def iendswith(self, other, **options):
        """
        implements the `endswith` operator.

        produces an `ilike` expression that tests against a match for the
        end of a string value. for example:
        column like '%' || <other> or column like '_' || <other>
        this method provides a case-insensitive variant of `endswith()` method.

        :param str other: expression to be compared.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword int start_count: count of `_` chars to be attached to start.
                                  if not provided, `%` will be used.

        :note start_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                           if the provided value goes upper than this limit,
                           a `%` will be attached instead of it. this limit
                           is for security reason.
        """

        escape = options.pop('escape', None)
        options.update(exact_start=False)
        other = self._process_like_autoescape(other, escape, options.get('autoescape', False))
        other = self._process_like_prefix(other, **options)
        return self.ilike(other, escape)

    def icontains(self, other, **options):
        """
        implements the 'contains' operator.

        produces an `ilike` expression that tests against a match for the
        middle of a string value. for example:
        column like `%` || <other> || `%` or column like `_` || <other> || `_`
        this method provides a case-insensitive variant of `contains()` method.

        :param str other: expression to be compared.

        :keyword str escape: optional escape character,
                             renders the `escape` keyword.

        :keyword bool autoescape: establishes an escape character within the like
                                  expression, then applies it to all occurrences of
                                  `%`, `_` and the escape character itself within the
                                  comparison value.

        :keyword int start_count: count of `_` chars to be attached to beginning.
                                  if not provided, `%` will be used.

        :keyword int end_count: count of `_` chars to be attached to end.
                                if not provided, `%` will be used.

        :note start_count, end_count: this value has a limit of `LIKE_CHAR_COUNT_LIMIT`,
                                      if the provided value goes upper than this limit,
                                      a `%` will be attached instead of it. this limit
                                      is for security reason.
        """

        escape = options.pop('escape', None)
        options.update(exact_start=False, exact_end=False)
        other = self._process_like_autoescape(other, escape, options.get('autoescape', False))
        other = self._process_like_prefix(other, **options)
        other = self._process_like_suffix(other, **options)
        return self.ilike(other, escape)

    def in_(self, other):
        """
        implements the `in` operator.

        in a column context, produces the clause `column in other`.

        the given parameter `other` may be:

        * a list of literal values.
        * a list of tuples may be provided if the comparison is against a
          `.tuple_` containing multiple expressions.
        * an empty list.
        * a bound parameter, for example `.bindparam`, may be used if it
          includes the `.bindparam.expanding` flag.
        * a :func:`_expression.select` construct, which is usually a correlated
          scalar select.
        * a single literal value could also be set and it will be converted into iterable.

        :param list | object | BindParameter | Select other: a list of literals, a
                                                             `_expression.select` construct,
                                                             or a `.bindparam` construct that
                                                             includes the `.bindparam.expanding`
                                                             flag set to True.
        """

        if not isinstance(other, (list, tuple, set, Select, BindParameter)):
            other = misc_utils.make_iterable(other, list)

        return super().in_(other)



# -*- coding: utf-8 -*-
"""
request wrappers base module.
"""

from collections import OrderedDict

from flask import Request

import pyrin.utils.unique_id as uuid_utils
import pyrin.globalization.datetime.services as datetime_services
import pyrin.globalization.locale.services as locale_services
import pyrin.converters.deserializer.services as deserializer_services
import pyrin.configuration.services as config_services
import pyrin.database.paging.services as paging_services

from pyrin.core.globals import _
from pyrin.caching.structs import CacheableDict
from pyrin.core.enumerations import HTTPMethodEnum
from pyrin.caching.decorators import cached_property
from pyrin.core.structs import DTO, CoreImmutableMultiDict
from pyrin.processor.request.wrappers.structs import RequestContext
from pyrin.settings.static import APPLICATION_ENCODING, DEFAULT_COMPONENT_KEY
from pyrin.processor.exceptions import RequestUserAlreadySetError
from pyrin.processor.request.wrappers.exceptions import InvalidRequestContextKeyNameError, \
    RequestContextKeyIsAlreadyPresentError, BadRequestError, RequestDeserializationError, \
    JSONBodyDecodingError, BodyDecodingError, RequestComponentCustomKeyAlreadySetError, \
    LargeContentError


class CoreRequest(Request):
    """
    core request class.

    this class should be used for server request.
    """

    # charset of the request.
    charset = APPLICATION_ENCODING

    # class to be used as request context holder.
    request_context_class = RequestContext

    # class to be used for 'args' and 'form' data.
    parameter_storage_class = CoreImmutableMultiDict

    # class to be used for dict values from the incoming WSGI.
    dict_storage_class = CoreImmutableMultiDict

    # these are query param names that application expects for locale and timezone.
    LOCALE_PARAM_NAME = 'lang'
    TIMEZONE_PARAM_NAME = 'tz'

    # application expects an authorization header in request with this key.
    AUTHORIZATION_HEADER_KEY = 'Authorization'

    # application expects one of these keys in environ dict to get client ip from.
    CLIENT_IP_ENVIRON_KEY_1 = 'HTTP_X_REAL_IP'
    CLIENT_IP_ENVIRON_KEY_2 = 'REMOTE_ADDR'

    # application will store authorization header in request context with this key.
    AUTHORIZATION_CONTEXT_KEY = 'authorization'

    def __init__(self, environ, populate_request=True,
                 shallow=False, **options):
        """
        initializes an instance of CoreRequest.

        :param dict environ: a wsgi environment object.
        :param bool populate_request: specifies that request should be populated.
        :param bool shallow: specifies that request should not consume form data.
        """

        super().__init__(environ, populate_request, shallow)

        self._request_id = uuid_utils.generate_uuid4()
        self._request_date = datetime_services.now()
        self._user = None
        self._cacheable_user = None
        self._component_custom_key = DEFAULT_COMPONENT_KEY
        self._client_ip = self._get_client_ip()
        self._safe_content_length = self._get_safe_content_length()
        self._context = self.request_context_class()

        self._extract_authorization_header()

        # when an attempt is done to parse inputs, this will be set to True
        # to prevent retrying.
        self.__inputs_parsed = False

        # this holds the inputs of request. this value will be
        # calculated once per each request and cached.
        self._inputs = DTO()

        # this holds the query strings of request without paging and globalization params.
        self._query_strings = None

        # this holds all query strings of request including globalization params.
        self._all_query_strings = None

        # this holds the paging query strings if they were provided.
        self._paging_params = None

        # this hold the locale and timezone query strings if they were provided.
        self._globalization_params = None

        # this holds any exception that might be raised on json decoding.
        # when silent is not passed to 'get_inputs' method, if this value
        # is not None, it will be raised.
        self._json_decoding_error = None

        # this holds any exception that might be raised on query strings or
        # form data deserialization. when silent is not passed to 'get_inputs'
        # method, if this value is not None, it will be raised.
        self._deserialization_error = None

        # this holds any exception that might be raised on custom body decoding.
        # when silent is not passed to 'get_inputs' method, if this value
        # is not None, it will be raised.
        self._body_decoding_error = None

    def __str__(self):
        result = 'request id: "{request_id}", request date: "{request_date}", ' \
                 'user: "{user}", method: "{method}", route: "{route}", ' \
                 'endpoint: "{endpoint}", client_ip: "{client_ip}", locale: "{locale}", ' \
                 'timezone: "{timezone}", component_custom_key: "{component}"'
        return result.format(request_id=self.request_id,
                             request_date=self.request_date,
                             user=self.user,
                             client_ip=self.client_ip,
                             route=self.path,
                             method=self.method,
                             endpoint=self._get_endpoint(),
                             locale=self.locale,
                             timezone=self.timezone.zone,
                             component=self.component_custom_key)

    def __hash__(self):
        return hash(self.request_id)

    def _get_endpoint(self):
        """
        gets the endpoint if available.

        otherwise returns None.

        :rtype: str
        """

        if self.url_rule is not None:
            return self.url_rule.endpoint

        return None

    def _get_client_ip(self):
        """
        gets client ip from environ if available, otherwise returns None.

        :rtype: str
        """

        return self.environ.get(self.CLIENT_IP_ENVIRON_KEY_1,
                                self.environ.get(self.CLIENT_IP_ENVIRON_KEY_2, None))

    def _extract_authorization_header(self):
        """
        extracts authorization header if available and puts it into
        request context, otherwise puts None into request context.
        """

        self.add_context(self.AUTHORIZATION_CONTEXT_KEY,
                         self.headers.get(self.AUTHORIZATION_HEADER_KEY, None))

    def _remove_extra_query_params(self, params):
        """
        removes all kwargs that should not be handed to the view function directly.

        for example `LOCALE_PARAM_NAME` and `TIMEZONE_PARAM_NAME` will be removed
        from query params because they will be stored in request object.
        the paging parameters will also be removed and stored in `_paging_params`
        dict attribute. this method removes extra kwargs from input dict directly
        and does not return anything.

        :param dict params: a dict containing all query params.
        """

        locale = params.pop(self.LOCALE_PARAM_NAME, None)
        timezone = params.pop(self.TIMEZONE_PARAM_NAME, None)
        globalization = dict()
        if locale not in (None, ''):
            globalization[self.LOCALE_PARAM_NAME] = locale

        if timezone not in (None, ''):
            globalization[self.TIMEZONE_PARAM_NAME] = timezone

        self._globalization_params = globalization
        self._paging_params = paging_services.extract_paging_params(params)

    def _get_safe_content_length(self):
        """
        gets content bytes length of this request if available, otherwise returns 0.

        :rtype: int
        """

        return self.content_length or 0

    def _deserialize(self, value, silent=False):
        """
        deserializes the given dict and returns the result dict.

        :param dict value: value to be deserialized.

        :param bool silent: specifies that if an error occurred on processing
                            the value, it should not raise an error.
                            defaults to False.

        :raises RequestDeserializationError: request deserialization error.

        :rtype: dict
        """

        try:
            return deserializer_services.deserialize(value, include_internal=False)
        except Exception as error:
            if silent is not True:
                self.on_deserialization_failed(error=error)
            self._deserialization_error = error
            return {}

    def get_body(self, silent=False):
        """
        gets data from request body.

        this method gets body as a dict if it is a json data.
        otherwise it returns an empty dict.

        :param bool silent: specifies that if an error occurred on processing
                            request body, it should not raise an error.
                            defaults to False.

        :raises BadRequestError: bad request error.

        :rtype: dict
        """

        if self.is_json is True:
            try:
                return self.get_json(silent=False) or {}
            except Exception as json_error:
                if silent is not True:
                    raise
                self._json_decoding_error = json_error
                return {}

        try:
            return self._get_custom_body() or {}
        except Exception as body_error:
            if silent is not True:
                self.on_body_loading_failed(error=body_error)
            self._body_decoding_error = body_error
            return {}

    def _get_custom_body(self):
        """
        gets data from a custom format request body.

        this method is intended to be overridden by subclasses.
        they must return an empty dict if could not convert the body.

        :rtype: dict
        """

        return {}

    def get_inputs(self, silent=False):
        """
        gets request inputs.

        not that if the same keyword is available in different param
        holders, the value will be replaced and no error will be raised.
        the replacement priority is as follows:
        [query_strings] -> [body or form_data] -> [view_args] -> [uploaded_files]

        for example, if a keyword named 'sample_key' is available in both body and
        query strings, the value of body will be available at the end.

        note that in the context of http requests, it is not possible for a
        request to have both body and form data at the same time. so this method
        will raise an error in such scenario.

        :param bool silent: specifies that if an error occurred on processing
                            request, it should not raise an error.
                            defaults to False.
                            note that if content length is above the limit, and
                            `silent=True` is given, inputs will not be parsed and
                            this method returns an empty dict.

        :raises LargeContentError: large content error.
        :raises JSONBodyDecodingError: json body decoding error.
        :raises RequestDeserializationError: request deserialization error.
        :raises BodyDecodingError: body decoding error.

        :rtype: dict
        """

        if self.url_rule is None:
            return DTO()

        if self.safe_content_length > self.url_rule.max_content_length:
            if silent is not True:
                raise LargeContentError(_('Request content is too large.'))

            return DTO()

        if silent is not True:
            if self._json_decoding_error is not None:
                raise self._json_decoding_error

            if self._deserialization_error is not None:
                self.on_deserialization_failed(error=self._deserialization_error)

            if self._body_decoding_error is not None:
                self.on_body_loading_failed(error=self._body_decoding_error)

        if self.__inputs_parsed is False:
            self.__inputs_parsed = True
            form_data = self._deserialize(self.form.to_dict(flat=False, all_list=False),
                                          silent=silent)

            body = self.get_body(silent=silent)

            self._inputs = DTO(self.get_query_strings(silent=silent))
            self._inputs.update(body)
            self._inputs.update(form_data)
            self._inputs.update(self.view_args or {})

            if self.files is not None and len(self.files) > 0:
                self._inputs.update(uploaded_files=self.files.to_dict(flat=False,
                                                                      all_list=False))

        return self._inputs

    def get_query_strings(self, silent=False):
        """
        gets the dict of query strings of current request.

        the result does not include the globalization query strings.

        :param bool silent: specifies that if an error occurred on processing
                            request query strings, it should not raise an error.
                            defaults to False.

        :raises RequestDeserializationError: request deserialization error.

        :rtype: dict
        """

        if self._query_strings is None:
            query_strings = self._deserialize(self.args.to_dict(flat=False, all_list=False),
                                              silent=silent)
            self._remove_extra_query_params(query_strings)
            self._query_strings = DTO(query_strings)

        return self._query_strings

    def get_all_query_strings(self, silent=False):
        """
        gets the dict of all query strings of current request.

        the result also includes the globalization query strings if they were provided.

        :param bool silent: specifies that if an error occurred on processing
                            request query strings, it should not raise an error.
                            defaults to False.

        :raises RequestDeserializationError: request deserialization error.

        :rtype: dict
        """

        if self._all_query_strings is None:
            self.get_query_strings(silent=silent)

            result = OrderedDict()
            result.update(self._globalization_params)
            result.update(self._query_strings)
            self._all_query_strings = result

        return self._all_query_strings

    def get_paging_params(self):
        """
        gets the paging params of current request.

        :returns: dict(int page: page number,
                       int page_size: page size)
        :rtype: dict
        """

        if self._paging_params is None:
            self.get_query_strings(silent=True)

        return self._paging_params

    def add_context(self, key, value, **options):
        """
        adds the given key/value pair into current request context.

        :param str key: key name to be added.
        :param object value: value to be added.

        :keyword bool replace: specifies that if a key with the same name
                               is already present, replace it. otherwise
                               raise an error. defaults to False if not provided.

        :raises InvalidRequestContextKeyNameError: invalid request context key name error.
        :raises RequestContextKeyIsAlreadyPresentError: request context key is
                                                        already present error.
        """

        if key in (None, '') or key.isspace():
            raise InvalidRequestContextKeyNameError('Request context key must be provided.')

        if key in self._context:
            replace = options.get('replace', None)
            if replace is None:
                replace = False

            if replace is not True:
                raise RequestContextKeyIsAlreadyPresentError('A request context with key '
                                                             '[{key}] is already present '
                                                             'and "replace" option is not set.'
                                                             .format(key=key))
        self._context[key] = value

    def get_context(self, key, default=None):
        """
        gets the value for given key from current request context.

        it gets the default value if key is not present in the request context.

        :param str key: key name to get its value.
        :param object default: a value to be returned if the provided
                               key is not present in request context.

        :rtype: object
        """

        return self._context.get(key, default)

    def remove_context(self, key):
        """
        removes the specified key from current request context if available.

        :param str key: key name to be removed from request context.
        """

        self._context.pop(key, None)

    def on_json_loading_failed(self, error):
        """
        raises an error on json loading failure.

        :param Exception error: exception that occurred on loading the json body.

        :raises JSONBodyDecodingError: json body decoding error.
        """

        if config_services.get_active('environment', 'debug', default=False) is True:
            raise JSONBodyDecodingError('Failed to decode JSON '
                                        'object: [{error}]'.format(error=error))

        self._on_bad_request_received(JSONBodyDecodingError)

    def on_body_loading_failed(self, error):
        """
        raises an error on body loading failure.

        :param Exception error: exception that occurred on loading the body.

        :raises BodyDecodingError: body decoding error.
        """

        if config_services.get_active('environment', 'debug', default=False) is True:
            raise BodyDecodingError('Failed to decode request '
                                    'body: [{error}]'.format(error=error))

        self._on_bad_request_received(BodyDecodingError)

    def on_deserialization_failed(self, error):
        """
        raises an error on query strings or form data deserialization failure.

        :param Exception error: exception that occurred on deserialization.

        :raises RequestDeserializationError: request deserialization error.
        """

        if config_services.get_active('environment', 'debug', default=False) is True:
            raise RequestDeserializationError('Failed to deserialize request '
                                              'query strings or form data: [{error}]'
                                              .format(error=error))

        self._on_bad_request_received(RequestDeserializationError)

    def _on_bad_request_received(self, error_class):
        """
        raises a generic error on receiving a bad request.

        :param type[BadRequestError] error_class: error class type that has been occurred.

        :raises BadRequestError: bad request error.
        """

        raise error_class(_('The browser (or proxy) sent a request '
                            'that this server could not understand.'))

    @property
    def request_id(self):
        """
        gets current request's request id.

        :rtype: uuid.UUID
        """

        return self._request_id

    @property
    def request_date(self):
        """
        gets current request's request date.

        :rtype: datetime
        """

        return self._request_date

    @property
    def user(self):
        """
        gets current request's user.

        :rtype: object
        """

        return self._user

    @user.setter
    def user(self, user):
        """
        sets current request's user.

        :param object user: user to be set.

        :raises RequestUserAlreadySetError: request user already set error.
        """

        if self._user is not None:
            raise RequestUserAlreadySetError('Request user for current request '
                                             'has been already set.')

        self._user = user

        if isinstance(user, dict):
            self._cacheable_user = CacheableDict(user)
        else:
            self._cacheable_user = user

    @property
    def cacheable_user(self):
        """
        gets the cacheable object of current user.

        if the user object is not a dict, it returns the same object.
        otherwise returns a `CacheableDict` object of user dict.

        :rtype: object | CacheableDict
        """

        return self._cacheable_user

    @property
    def component_custom_key(self):
        """
        gets current request's component custom key.

        :rtype: object
        """

        return self._component_custom_key

    @component_custom_key.setter
    def component_custom_key(self, component_custom_key):
        """
        sets current request's component custom key.

        :param object component_custom_key: component custom key to be set.

        :raises RequestComponentCustomKeyAlreadySetError: request component custom
                                                          key already set error.
        """

        if self._component_custom_key != DEFAULT_COMPONENT_KEY:
            raise RequestComponentCustomKeyAlreadySetError('Request component custom key '
                                                           'for current request has been '
                                                           'already set.')

        self._component_custom_key = component_custom_key

    @property
    def client_ip(self):
        """
        gets current request's client ip if available, otherwise returns None.

        :rtype: str
        """

        return self._client_ip

    @property
    def safe_content_length(self):
        """
        gets current request's content length if available, otherwise returns 0.

        :rtype: int
        """

        return self._safe_content_length

    @cached_property
    def locale(self):
        """
        gets the locale of current request.

        :rtype: str
        """

        locale = self.args.get(self.LOCALE_PARAM_NAME, None)
        if locale not in (None, '') and locale_services.locale_exists(locale) is True:
            return locale

        return locale_services.get_default_locale()

    @cached_property
    def timezone(self):
        """
        gets the timezone of current request.

        :rtype: tzinfo
        """

        timezone_name = self.args.get(self.TIMEZONE_PARAM_NAME, None)
        if timezone_name not in (None, ''):
            try:
                return datetime_services.get_timezone(timezone_name)
            except Exception:
                pass

        return datetime_services.get_default_client_timezone()

    @property
    def authorization(self):
        """
        gets the authorization header of current request.

        returns None if authorization header is not set.

        :rtype: str
        """

        return self.get_context(self.AUTHORIZATION_CONTEXT_KEY, None)

    @property
    def is_preflight(self):
        """
        gets a value indicating that this is a preflight request.

        it returns True if request method is `OPTIONS` and two required
        preflight headers are present. which are `Access-Control-Request-Method`
        and `Origin`. otherwise returns False.

        :rtype: bool
        """

        if self.method != HTTPMethodEnum.OPTIONS or self.origin in (None, ''):
            return False

        if self.access_control_request_method in (None, ''):
            return False

        return True

    @property
    def is_cors(self):
        """
        gets a value indicating that this is a cors request.

        it returns True if request method is not `OPTIONS` and `Origin`
        header is present, otherwise returns False.

        :rtype: bool
        """

        return self.origin not in (None, '') and self.method != HTTPMethodEnum.OPTIONS



# -*- coding: utf-8 -*-
"""
babel package.
"""

from pyrin.packaging.base import Package


class BabelPackage(Package):
    """
    babel package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'globalization.locale.babel.component'
    CONFIG_STORE_NAMES = ['babel']
    EXTRA_CONFIG_STORE_NAMES = ['babel.mappings']
    DEPENDS = ['pyrin.configuration',
               'pyrin.cli']



# -*- coding: utf-8 -*-
"""
mimetype handlers html module.
"""

import re

from pyrin.processor.mimetype.decorators import mimetype_handler
from pyrin.processor.mimetype.enumerations import MIMETypeEnum
from pyrin.processor.mimetype.handlers.base import MIMETypeHandlerBase


@mimetype_handler()
class HTMLMIMETypeHandler(MIMETypeHandlerBase):
    """
    html mimetype handler class.

    this class detects html markup in most inclusive way. it's not a
    real html parser. it just provides a way to detect if a string
    has potential html markup inside it.
    """

    # matches the html inside string.
    # matching is case-insensitive.
    HTML_REGEX = re.compile(r'^.*<.*\S+.*>.*', re.IGNORECASE | re.MULTILINE | re.DOTALL)

    def __init__(self, **options):
        """
        creates an instance of HTMLMIMETypeHandler.
        """

        super().__init__(**options)

    def _mimetype(self, value, **options):
        """
        gets the mimetype of given value.

        returns None if it fails to detect the mimetype.

        :param str value: value to detect its mimetype.

        :rtype: str
        """

        if self.HTML_REGEX.match(value.strip()):
            return MIMETypeEnum.HTML

        return None

    @property
    def accepted_type(self):
        """
        gets the accepted type for this mimetype handler.

        which could detect mimetype from this type.

        :rtype: type[str]
        """

        return str



# -*- coding: utf-8 -*-
"""
sentry component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.logging.sentry import SentryPackage
from pyrin.logging.sentry.manager import SentryManager


@component(SentryPackage.COMPONENT_NAME)
class SentryComponent(Component, SentryManager):
    """
    sentry component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
validator handlers exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException
from pyrin.validator.exceptions import ValidationError


class ValidatorHandlersException(CoreException):
    """
    validator handlers exception.
    """
    pass


class ValidatorHandlersBusinessException(CoreBusinessException,
                                         ValidatorHandlersException):
    """
    validator handlers business exception.
    """
    pass


class HandlersValidationException(ValidationError,
                                  ValidatorHandlersBusinessException):
    """
    handlers validation exception.

    this class should be used as the base class for all handlers validation errors.
    """
    pass


class ValidatorFieldIsRequiredError(ValidatorHandlersException):
    """
    validator field is required error.
    """
    pass


class ValueCouldNotBeNoneError(HandlersValidationException):
    """
    value could not be none error.
    """
    pass


class ValueIsNotListError(HandlersValidationException):
    """
    value is not list error.
    """
    pass


class ValueCouldNotBeAnEmptyListError(HandlersValidationException):
    """
    value could not be an empty list error.
    """
    pass


class ValueIsLowerThanMinimumError(HandlersValidationException):
    """
    value is lower than minimum error.
    """
    pass


class ValueIsHigherThanMaximumError(HandlersValidationException):
    """
    value is higher than maximum error.
    """
    pass


class ValueIsOutOfRangeError(HandlersValidationException):
    """
    value is out of range error.
    """
    pass


class InvalidValueTypeError(HandlersValidationException):
    """
    invalid value type error.
    """
    pass


class ValueDoesNotMatchPatternError(HandlersValidationException):
    """
    value does not match pattern error.
    """
    pass


class InvalidValidatorDomainError(ValidatorHandlersException):
    """
    invalid validator domain error.
    """
    pass


class InvalidAcceptedTypeError(ValidatorHandlersException):
    """
    invalid accepted type error.
    """
    pass


class ValidatorFixerMustBeCallable(ValidatorHandlersException):
    """
    validator fixer must be callable.
    """
    pass


class InvalidValidationExceptionTypeError(ValidatorHandlersException):
    """
    invalid validation exception type error.
    """
    pass


class AcceptedMinimumValueMustBeProvidedError(ValidatorHandlersException):
    """
    accepted minimum value must be provided error.
    """
    pass


class AcceptedMaximumValueMustBeProvidedError(ValidatorHandlersException):
    """
    accepted maximum value must be provided error.
    """
    pass


class ValidValuesMustBeProvidedError(ValidatorHandlersException):
    """
    valid values must be provided error.
    """
    pass


class InvalidValuesMustBeProvidedError(ValidatorHandlersException):
    """
    invalid values must be provided error.
    """
    pass


class LongStringLengthError(HandlersValidationException):
    """
    long string length error.
    """
    pass


class ShortStringLengthError(HandlersValidationException):
    """
    short string length error.
    """
    pass


class ValueCouldNotBeBlankError(HandlersValidationException):
    """
    value could not be blank error.
    """
    pass


class ValueCouldNotBeWhitespaceError(HandlersValidationException):
    """
    value could not be whitespace error.
    """
    pass


class InvalidRegularExpressionError(ValidatorHandlersException):
    """
    invalid regular expression error.
    """
    pass


class RegularExpressionMustBeProvidedError(ValidatorHandlersException):
    """
    regular expression must be provided error.
    """
    pass


class ValueIsNotNumberError(InvalidValueTypeError):
    """
    value is not number error.
    """
    pass


class ValueIsNotIntegerError(InvalidValueTypeError):
    """
    value is not integer error.
    """
    pass


class ValueIsNotFloatError(InvalidValueTypeError):
    """
    value is not float error.
    """
    pass


class ValueIsNotDecimalError(InvalidValueTypeError):
    """
    value is not decimal error.
    """
    pass


class ValueIsNotStringError(InvalidValueTypeError):
    """
    value is not string error.
    """
    pass


class ValueIsNotDateTimeError(InvalidValueTypeError):
    """
    value is not datetime error.
    """
    pass


class ValueIsNotDateError(InvalidValueTypeError):
    """
    value is not date error.
    """
    pass


class ValueIsNotTimeError(InvalidValueTypeError):
    """
    value is not time error.
    """
    pass


class ValueIsNotBooleanError(InvalidValueTypeError):
    """
    value is not boolean error.
    """
    pass


class ValueIsNotUUIDError(InvalidValueTypeError):
    """
    value is not uuid error.
    """
    pass


class ValueIsNotDictError(InvalidValueTypeError):
    """
    value is not dict error.
    """
    pass


class ValueIsNotUUID4Error(InvalidValueTypeError):
    """
    value is not uuid4 error.
    """
    pass


class MinimumLengthHigherThanMaximumLengthError(ValidatorHandlersException):
    """
    minimum length higher than maximum length error.
    """
    pass


class MinimumValueLargerThanMaximumValueError(ValidatorHandlersException):
    """
    minimum value larger than maximum value error.
    """
    pass


class InvalidEmailError(HandlersValidationException):
    """
    invalid email error.
    """
    pass


class InvalidIPv4Error(HandlersValidationException):
    """
    invalid ipv4 error.
    """
    pass


class InvalidURLError(HandlersValidationException):
    """
    invalid url error.
    """
    pass


class InvalidHTTPURLError(InvalidURLError):
    """
    invalid http url error.
    """
    pass


class InvalidHTTPSURLError(InvalidURLError):
    """
    invalid https url error.
    """
    pass



# -*- coding: utf-8 -*-
"""
orm sql schema exceptions module.
"""

from pyrin.core.exceptions import CoreException


class ORMSQLSchemaException(CoreException):
    """
    orm sql schema exception.
    """
    pass


class SequencePKColumnTypeIsInvalidError(ORMSQLSchemaException):
    """
    sequence pk column type is invalid error.
    """
    pass


class AutoPKColumnTypeIsInvalidError(ORMSQLSchemaException):
    """
    auto pk column type is invalid error.
    """
    pass


class InvalidFKColumnReferenceTypeError(ORMSQLSchemaException):
    """
    invalid fk column reference type error.
    """
    pass


class CheckConstraintConflictError(ORMSQLSchemaException):
    """
    check constraint conflict error.
    """
    pass


class InvalidCheckConstraintError(ORMSQLSchemaException):
    """
    invalid check constraint error.
    """
    pass


class StringColumnTypeIsInvalidError(ORMSQLSchemaException):
    """
    string column type is invalid error.
    """
    pass



# -*- coding: utf-8 -*-
"""
packaging package.
"""

from pyrin.packaging.base import Package


class PackagingPackage(Package):
    """
    packaging package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'packaging.component'
    CONFIG_STORE_NAMES = ['packaging']



# -*- coding: utf-8 -*-
"""
database manager module.
"""

from sqlalchemy import engine_from_config

import pyrin.configuration.services as config_services
import pyrin.processor.response.status.services as status_services
import pyrin.logging.services as logging_services
import pyrin.security.session.services as session_services
import pyrin.processor.response.services as response_services
import pyrin.utils.dictionary as dict_utils
import pyrin.utils.misc as misc_utils

from pyrin.database import DatabasePackage
from pyrin.database.hooks import DatabaseHookBase
from pyrin.core.mixin import HookMixin
from pyrin.database.model.base import BaseEntity
from pyrin.core.structs import DTO, Manager
from pyrin.core.enumerations import ServerErrorResponseCodeEnum
from pyrin.database.interface import AbstractSessionFactoryBase
from pyrin.utils.custom_print import print_warning
from pyrin.database.exceptions import InvalidSessionFactoryTypeError, \
    DuplicatedSessionFactoryError, SessionFactoryNotExistedError, InvalidEntityTypeError, \
    InvalidDatabaseBindError, InvalidDatabaseHookTypeError


class DatabaseManager(Manager, HookMixin):
    """
    database manager class.
    """

    LOGGER = logging_services.get_logger('database')
    BIND_REMOVE_KEY_PREFIX = '__'
    DEFAULT_DATABASE_NAME = 'default'
    hook_type = DatabaseHookBase
    invalid_hook_type_error = InvalidDatabaseHookTypeError
    package_class = DatabasePackage

    def __init__(self):
        """
        initializes an instance of DatabaseManager.
        """

        super().__init__()

        # contains the application default database engine.
        self.__engine = self._create_default_engine()

        # a dictionary containing engines for different bounded databases.
        # in the form of: {str bind_name: Engine engine}
        self.__bounded_engines = self._create_bounded_engines()

        # a dictionary containing all entity classes that should be
        # bounded into a database engine other than the default one.
        # in the form of: {type entity: str bind_name}
        self._binds = DTO()

        # a dictionary containing entity to engine map for those entities that
        # should be bounded to a different database engine than the default one.
        # in the form of: {type entity: Engine engine}
        self._entity_to_engine_map = DTO()

        # a dictionary containing table name to engine map for those entities that
        # should be bounded to a different database engine than the default one.
        # in the form of: {string table_name: Engine engine}
        self._table_name_to_engine_map = DTO()

        # a dictionary containing session factories for request bounded and unbounded types.
        # in the for of: {bool request_bounded: Session session_factory}
        # it should have at most two different keys, True for request bounded
        # and False for request unbounded.
        self._session_factories = DTO()

    def get_current_store(self, **kwargs):
        """
        gets current database store.

        note that this method will always get the correct session based on available
        context. so if you are in an atomic context, it gets you the correct atomic
        session, but if you are not in an atomic context, it will get you the related
        session to current scope.

        :keyword object **kwargs: keyword arguments will be passed to the
                                  `CoreScopedSession.session_factory` callable
                                  to configure the new session that's being
                                  created, if an existing session is not present.
                                  if the session is present and keyword arguments
                                  have been passed, an error will be raised.

        :raises ScopedSessionIsAlreadyPresentError: scoped session is already present error.

        :returns: database session
        :rtype: CoreSession
        """

        return self._get_current_session_factory()(**kwargs)

    def get_current_session_factory(self):
        """
        gets current database session factory.

        this method should not be used directly for data manipulation.
        use `get_current_store` method instead.

        :returns: database session factory
        :rtype: CoreScopedSession
        """

        return self._get_current_session_factory()

    def get_atomic_store(self, **kwargs):
        """
        gets an atomic database store.

        atomic is meaning a new session with a new transaction. note that it's normally
        not needed to get an atomic session manually, instead you could use `@atomic`
        decorator to provide you an atomic session. but if you really need to get an atomic
        session manually, you have to manually remove that session from corresponding
        session factory after you've done. otherwise, unexpected behaviour may occur.
        so if you get an atomic session, and don't remove it, after that if you get another
        session in the same scope, you will get the same exact atomic session. but if you
        get an atomic session, and remove it from corresponding session factory after you've
        done, after that if you get a session, it will get you a session related to current
        scope. this is why it's recommended not to get an atomic session manually, and
        instead use `@atomic` decorator when you need an atomic session.

        :keyword object **kwargs: keyword arguments will be passed to the
                                  `CoreScopedSession.session_factory` callable
                                  to configure the new atomic session that's
                                  being created.

        :returns: database session
        :rtype: CoreSession
        """

        return self._get_current_session_factory()(atomic=True, **kwargs)

    def _get_current_session_factory(self):
        """
        gets database session factory that should be used.

        it simply checks that request is available in current
        context or not, and gets the correct session factory.

        :returns: database session factory
        :rtype: CoreScopedSession
        """

        return self._get_session_factory(session_services.is_request_context_available())

    def _get_session_factory(self, request_bounded):
        """
        gets the session factory based on given input.

        :param bool request_bounded: a value indicating that the session
                                     factory should be bounded into request.

        :raises SessionFactoryNotExistedError: session factory not existed error.

        :returns: database session factory
        :rtype: CoreScopedSession
        """

        if request_bounded not in self._session_factories:
            raise SessionFactoryNotExistedError('Session factory with '
                                                'request_bounded={bounded} '
                                                'is not available.'
                                                .format(bounded=request_bounded))

        return self._session_factories.get(request_bounded)

    def _create_default_engine(self):
        """
        creates and gets the default database engine using database configuration store.

        :returns: database engine
        :rtype: Engine
        """

        database_configs = config_services.get_active_section('database')
        return self._create_engine(database_configs)

    def _create_engine(self, database_configs, **kwargs):
        """
        creates a database engine using specified database configuration and returns it.

        each provided key from kwargs will override the corresponding
        key in database_configs dict. note that kwargs should not have any prefix.

        :param dict database_configs: database configs that should be used.

        :returns: database engine
        :rtype: Engine
        """

        configs_prefix = self.get_configs_prefix()
        return engine_from_config(database_configs, prefix=configs_prefix, **kwargs)

    def get_configs_prefix(self):
        """
        gets the configs prefix for sqlalchemy keys in database config store.

        it gets the value of `configs_prefix` key from database config store.

        :rtype: str
        """

        return config_services.get_active('database', 'configs_prefix')

    def _create_bounded_engines(self):
        """
        creates all required bounded engines if any and returns them.

        :returns: dict[str bind_name: Engine engine]
        :rtype: dict
        """

        engines = DTO()
        base_database_configs = config_services.get_active_section('database')
        binds_names = base_database_configs['bind_names']

        if binds_names is None:
            return engines
        binds_names = misc_utils.make_iterable(binds_names, list)
        if len(binds_names) <= 0:
            return engines

        for name in binds_names:
            bind_configs = self._get_bind_configs(name)
            full_configs = self._merge_configs(base_database_configs, bind_configs)
            engine = self._create_engine(full_configs)
            engines[name] = engine

        return engines

    def _merge_configs(self, base_configs, bind_configs):
        """
        merges given base and bind configs and returns a new dict.

        merging will cause all keys in bind configs override any available key in
        base configs. also all keys that are present in bind configs and starting
        with double underscore `__` will be removed from result dict and also all
        keys that their names are exactly like prefixed ones but without `__` will
        be removed from result dict.

        :param dict base_configs: base configs from `database.ini` file.
        :param dict bind_configs: bind configs from `database.binds.ini`

        :rtype: dict
        """

        result = DTO(**base_configs)
        result.update(**bind_configs)
        result = dict_utils.remove_keys(result, self.BIND_REMOVE_KEY_PREFIX)

        return result

    def _get_bind_configs(self, bind_name):
        """
        gets all bind configs of given bind name for currently
        active environment in 'database.ini' from 'database.binds.ini' file.

        :param str bind_name: bind name to get its configs.

        :rtype: dict
        """

        bind_configs_section = self.get_bind_config_section_name(bind_name)
        return config_services.get_section('database.binds', bind_configs_section)

    def get_bind_config_section_name(self, bind_name):
        """
        gets the bind config section name for given bind name and currently
        active environment in 'database.ini' from 'database.binds.ini' file.

        :param str bind_name: bind name to get its section name.

        :rtype: str
        """

        active_environment = config_services.get_active_section_name('database')
        bind_config_key = '{environment}_{bind_name}'.format(environment=active_environment,
                                                             bind_name=bind_name)

        return bind_config_key

    def finalize_transaction(self, response, **options):
        """
        this method will finalize database transaction of each request.

        this method will finalize both normal and atomic sessions of
        current request if available.
        we should not raise any exception in finalize transaction hook,
        so we return an error response in case of any exception.
        note that normally you should never call this method manually.

        :param CoreResponse response: response object.

        :rtype: CoreResponse | tuple
        """

        try:
            session_factory = self.get_current_session_factory()
            try:
                if status_services.is_error(response.status_code,
                                            strict_status=False) is True:
                    session_factory.rollback_all()
                    return response

                session_factory.commit_all()
                return response
            except Exception:
                session_factory.rollback_all()
                raise
            finally:
                session_factory.remove()
        except Exception as error:
            self.LOGGER.exception(str(error))
            return response_services.make_exception_response(error,
                                                             code=ServerErrorResponseCodeEnum.
                                                             INTERNAL_SERVER_ERROR)

    def cleanup_session(self, exception):
        """
        this method will cleanup database session of each request.

        this method will cleanup both normal and atomic sessions of
        current request if available.
        in case of any unhandled exception. we should not raise any exception
        in teardown request handlers, so we just log the exception.
        note that normally you should never call this method manually.

        :param Exception exception: exception instance.
        """

        if exception is not None:
            try:
                self.LOGGER.exception(str(exception))
                session_factory = self.get_current_session_factory()
                session_factory.remove()

            except Exception as error:
                self.LOGGER.exception(str(error))

    def register_session_factory(self, instance, **options):
        """
        registers a new session factory or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding an instance which it's `request_bounded` is already available
        in registered session factories.

        :param AbstractSessionFactoryBase instance: session factory to be registered.
                                                    it must be an instance of
                                                    AbstractSessionFactoryBase.

        :keyword bool replace: specifies that if there is another registered
                               session factory with the same `request_bounded`,
                               replace it with the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidSessionFactoryTypeError: invalid session factory type error.
        :raises DuplicatedSessionFactoryError: duplicated session factory error.
        """

        if not isinstance(instance, AbstractSessionFactoryBase):
            raise InvalidSessionFactoryTypeError('Input parameter [{instance}] is '
                                                 'not an instance of [{base}].'
                                                 .format(instance=str(instance),
                                                         base=AbstractSessionFactoryBase))

        # checking whether is there any registered
        # instance with the same 'request_bounded' value.
        if instance.request_bounded in self._session_factories:
            replace = options.get('replace', False)

            if replace is not True:
                raise DuplicatedSessionFactoryError('There is another registered session factory '
                                                    'with "request_bounded={bounded}" but '
                                                    '"replace" option is not set, so session '
                                                    'factory [{instance}] could not be registered.'
                                                    .format(bounded=instance.request_bounded,
                                                            instance=str(instance)))

            old_instance = self._session_factories[instance.request_bounded]
            print_warning('Session factory [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=str(old_instance), new_instance=str(instance)))

        # registering new session factory.
        self._session_factories[instance.request_bounded] = \
            instance.create_session_factory(self.get_default_engine())

    def register_bind(self, entity, bind_name, **options):
        """
        binds the given model class with specified bind database.

        :param type[BaseEntity] entity: base entity subclass to be bounded.
        :param str bind_name: bind name to be associated with the model class.

        :raises InvalidEntityTypeError: invalid entity type error.
        """

        if not issubclass(entity, BaseEntity):
            raise InvalidEntityTypeError('Input parameter [{entity}] is '
                                         'not a subclass of [{base}].'
                                         .format(entity=entity,
                                                 base=BaseEntity))

        # registering model into database binds.
        self._binds[entity] = bind_name

    def _map_entity_to_engine(self):
        """
        maps all application entities that should be bounded to a
        database engine other than the default one to relevant engines.

        :raises InvalidDatabaseBindError: invalid database bind error.
        """

        binds = config_services.get_active('database', 'bind_names')
        binds = misc_utils.make_iterable(binds, list)

        for entity, bind_name in self._binds.items():
            if bind_name not in binds:
                raise InvalidDatabaseBindError('Database bind name [{bind_name}] for entity '
                                               '[{entity_name}] is not available in '
                                               'database config store.'
                                               .format(bind_name=bind_name,
                                                       entity_name=entity))

            self._entity_to_engine_map[entity] = self.get_bounded_engines()[bind_name]
            self._table_name_to_engine_map[entity.table_name.lower()] = \
                self.get_bounded_engines()[bind_name]
            self._table_name_to_engine_map[entity.table_fullname.lower()] = \
                self.get_bounded_engines()[bind_name]

    def configure_session_factories(self):
        """
        configures all application session factories.

        normally, you should not call this method manually.

        :raises InvalidDatabaseBindError: invalid database bind error.
        """

        self._map_entity_to_engine()
        if len(self.get_entity_to_engine_map()) > 0:
            for key in self._session_factories:
                self._session_factories[key].configure(binds=self.get_entity_to_engine_map())

        self._after_session_factories_configured()

    def get_default_engine(self):
        """
        gets database default engine.

        :rtype: Engine
        """

        return self.__engine

    def get_bounded_engines(self):
        """
        gets database bounded engines.

        :returns: dict[str bind_name: Engine engine]
        :rtype: dict
        """

        return self.__bounded_engines

    def get_entity_to_engine_map(self):
        """
        gets entity to engine map.

        :returns: dict[type entity, Engine engine]
        :rtype: dict
        """

        return self._entity_to_engine_map

    def get_table_name_to_engine_map(self):
        """
        gets table name to engine map.

        :returns: dict[str table_name, Engine engine]
        :rtype: dict
        """

        return self._table_name_to_engine_map

    def _after_session_factories_configured(self):
        """
        this will call `after_session_factories_configured` method of registered hooks.
        """

        for hook in self._get_hooks():
            hook.after_session_factories_configured()

    def get_entity_engine(self, entity):
        """
        gets the database engine which the provided entity class is bounded to.

        :param BaseEntity entity: entity class to get its bounded engine.

        :rtype: Engine
        """

        if entity in self.get_entity_to_engine_map():
            return self.get_entity_to_engine_map()[entity]

        return self.get_default_engine()

    def get_table_engine(self, table_name):
        """
        gets the database engine which the provided table name is bounded to.

        :param str table_name: table name to get its bounded engine.

        :rtype: Engine
        """

        table_name = table_name.lower()
        if table_name in self.get_table_name_to_engine_map():
            return self.get_table_name_to_engine_map()[table_name]

        return self.get_default_engine()

    def get_bind_name_engine(self, bind_name):
        """
        gets the database engine which the provided bind name is bounded to.

        :param str bind_name: bind name to get its bounded engine.

        :raises InvalidDatabaseBindError: invalid database bind error.

        :rtype: Engine
        """

        if bind_name == self.DEFAULT_DATABASE_NAME:
            return self.get_default_engine()

        elif bind_name in self.get_bounded_engines():
            return self.get_bounded_engines()[bind_name]

        raise InvalidDatabaseBindError('Database bind name [{bind_name}] does '
                                       'not exist in database config store.'
                                       .format(bind_name=bind_name))

    def get_default_database_name(self):
        """
        gets default database name.

        :rtype: str
        """

        return self.DEFAULT_DATABASE_NAME

    def get_all_database_names(self):
        """
        gets all database names defined in application.

        it returns all available database names, even those
        that do not have any entity bounded to them.
        the result also includes the default database name.

        :rtype: list[str]
        """

        result = [self.get_default_database_name()]
        binds = self.get_database_bind_names()
        result.extend(binds)

        return result

    def get_database_bind_names(self):
        """
        gets all database bind names defined in application.

        it returns all available database bind names, even those
        that do not have any entity bounded to them.

        :rtype: list[str]
        """

        result = []
        binds = config_services.get_active('database', 'bind_names')
        if binds is not None:
            result.extend(binds)

        return result



# -*- coding: utf-8 -*-
"""
cli services module.
"""

from pyrin.application.services import get_component
from pyrin.cli import CLIPackage


def register_cli_group(name, instance):
    """
    registers a cli group.

    :param str name: cli group name.
    :param CLI instance: cli group instance.

    :raises CLIGroupNameIsRequiredError: cli group name is required error.
    :raises DuplicatedCLIGroupError: duplicated cli group error.
    :raises InvalidCLIGroupTypeError: invalid cli group type error.
    """

    return get_component(CLIPackage.COMPONENT_NAME).register_cli_group(name, instance)


def get_cli_groups():
    """
    gets all registered cli groups.

    :rtype: CLIGroups
    """

    return get_component(CLIPackage.COMPONENT_NAME).get_cli_groups()


def process_function(func, func_args, func_kwargs):
    """
    processes the given cli handler function with given inputs.

    :param function func: function to be executed.
    :param tuple func_args: a tuple of function positional inputs.
    :param dict func_kwargs: a dictionary of function keyword arguments.

    :raises CLIHandlerNotFoundError: cli handler not found error.
    :raises InvalidCLIDecoratedMethodError: invalid cli decorated method error.

    :rtype: int
    """

    return get_component(CLIPackage.COMPONENT_NAME).process_function(func, func_args,
                                                                     func_kwargs)


def invoke_function(func, func_args, func_kwargs):
    """
    invokes the given function with given inputs.

    :param function func: function to be executed.
    :param tuple func_args: a tuple of function positional inputs.
    :param dict func_kwargs: a dictionary of function keyword arguments.

    :returns: function execution result.
    """

    return get_component(CLIPackage.COMPONENT_NAME).invoke_function(func, func_args,
                                                                    func_kwargs)



# -*- coding: utf-8 -*-
"""
celery cli handlers call module.
"""

from pyrin.task_queues.celery.cli.decorators import celery_cli_handler
from pyrin.task_queues.celery.cli.enumerations import CeleryCLIHandlersEnum
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase
from pyrin.task_queues.celery.cli.handlers.params import TaskNamePositionalParam, \
    ArgsParam, KwargsParam, ETAParam, CountdownParam, ExpiresParam, SerializerParam, \
    QueueParam, ExchangeParam, RoutingKeyParam


@celery_cli_handler()
class CallCLIHandler(CeleryCLIHandlerBase):
    """
    call cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of CallCLIHandler.
        """

        super().__init__(CeleryCLIHandlersEnum.CALL)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([TaskNamePositionalParam(index=0),
                       ArgsParam(), KwargsParam(),
                       ETAParam(), CountdownParam(),
                       ExpiresParam(), SerializerParam(),
                       QueueParam(), ExchangeParam(),
                       RoutingKeyParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
celery cli handlers amqp module.
"""

from pyrin.task_queues.celery.cli.decorators import celery_cli_handler
from pyrin.task_queues.celery.cli.enumerations import CeleryCLIHandlersEnum
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase


@celery_cli_handler()
class AMQPCLIHandler(CeleryCLIHandlerBase):
    """
    amqp cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of AMQPCLIHandler.
        """

        super().__init__(CeleryCLIHandlersEnum.AMQP)



# -*- coding: utf-8 -*-
"""
database migration exceptions module.
"""

from pyrin.core.exceptions import CoreException


class DatabaseMigrationManagerException(CoreException):
    """
    database migration manager exception.
    """
    pass


class EngineBindNameNotFoundError(DatabaseMigrationManagerException):
    """
    engine bind name not found error.
    """
    pass



# -*- coding: utf-8 -*-
"""
database audit package.
"""

from pyrin.packaging.base import Package


class DatabaseAuditPackage(Package):
    """
    database audit package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.audit.component'
    DEPENDS = ['pyrin.audit']



# -*- coding: utf-8 -*-
"""
api decorators module.
"""

import pyrin.api.services as api_services


def api_hook():
    """
    decorator to register an api hook.

    :raises InvalidAPIHookTypeError: invalid api hook type error.

    :returns: api hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available api hooks.

        :param type cls: api hook class.

        :returns: api hook class.
        :rtype: type
        """

        instance = cls()
        api_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
deserializer component module.
"""

from pyrin.converters.deserializer import DeserializerPackage
from pyrin.converters.deserializer.manager import DeserializerManager
from pyrin.application.decorators import component
from pyrin.application.structs import Component


@component(DeserializerPackage.COMPONENT_NAME)
class DeserializerComponent(Component, DeserializerManager):
    """
    deserializer component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
token exceptions module.
"""

from pyrin.core.exceptions import CoreException, CoreBusinessException


class TokenManagerException(CoreException):
    """
    token manager exception.
    """
    pass


class TokenManagerBusinessException(CoreBusinessException,
                                    TokenManagerException):
    """
    token manager business exception.
    """
    pass


class TokenDecodingError(TokenManagerBusinessException):
    """
    token decoding error.
    """
    pass


class TokenVerificationError(TokenDecodingError):
    """
    token verification error.
    """
    pass


class InvalidTokenHandlerTypeError(TokenManagerException):
    """
    invalid token handler type error.
    """
    pass


class DuplicatedTokenHandlerError(TokenManagerException):
    """
    duplicated token handler error.
    """
    pass


class TokenHandlerNotFoundError(TokenVerificationError):
    """
    token handler not found error.
    """
    pass


class InvalidTokenHandlerNameError(TokenManagerException):
    """
    invalid token handler name error.
    """
    pass


class TokenKidHeaderNotSpecifiedError(TokenVerificationError):
    """
    token kid header not specified error.
    """
    pass


class TokenKidHeaderNotFoundError(TokenVerificationError):
    """
    token kid header not found error.
    """
    pass


class DuplicatedTokenKidHeaderError(TokenManagerException):
    """
    duplicated token kid header error.
    """
    pass



# -*- coding: utf-8 -*-
"""
hs256 token handler module.
"""

import pyrin.configuration.services as config_services
import pyrin.security.utils.services as security_utils_services

from pyrin.security.token.decorators import token
from pyrin.security.token.handlers.base import SymmetricTokenBase


@token()
class HS256Token(SymmetricTokenBase):
    """
    hs256 token class.
    """

    def __init__(self, **options):
        """
        initializes an instance of HS256Token.
        """

        super().__init__(**options)

    def _get_encoding_key(self, **options):
        """
        gets the signing key for encoding.

        :rtype: str
        """

        return config_services.get('security', 'token', 'hs256_key')

    def _get_algorithm(self, **options):
        """
        gets the algorithm for signing the token.

        :rtype: str
        """

        return 'HS256'

    def generate_key(self, **options):
        """
        generates a valid key for this handler and returns it.

        :keyword int length: the length of generated key in bytes.
                             note that some token handlers may not accept custom
                             key length so this value would be ignored on those handlers.

        :rtype: str
        """

        key_length = options.get('length', None)
        if key_length is None:
            key_length = config_services.get('security', 'token', 'hs256_key_length')

        return security_utils_services.get_url_safe(length=key_length)

    def get_kid(self):
        """
        gets kid value to be used in token header for this handler.
        it must be unique for each handler.

        :rtype: str
        """

        return '494d8b2d-ae91-442d-86eb-9fd020fe3518'



# -*- coding: utf-8 -*-
"""
cli core template package.
"""

from pyrin.packaging.base import Package


class CLICoreTemplatePackage(Package):
    """
    cli core template package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
logging contexts module.
"""

import pyrin.logging.services as logging_services

from pyrin.core.contexts import ContextManagerBase


class suppress(ContextManagerBase):
    """
    context manager to suppress any exceptions and logs them.

    for example:

    def service1():
        with suppress():
            raise Exception('Error Occurred')

    def service2():
        with suppress(ValueError, TypeError, log=False):
            raise ValueError('Error Occurred')
    """

    def __init__(self, *errors, **options):
        """
        initializes an instance of suppress.

        :param type[Exception] errors: exception types to be suppressed.
                                       if not provided, all exceptions will
                                       be suppressed.

        :keyword bool log: log suppressed exceptions.
                           defaults to True if not provided.
        """

        if errors is None or len(errors) <= 0:
            errors = (Exception,)

        self._errors = errors
        self._log = options.get('log', True)

        super().__init__()

    def __exit__(self, exc_type, exc_value, traceback):
        """
        does the finalizing to suppress and log and required errors.

        :param type[Exception] exc_type: the exception type that has been
                                         occurred during current context.

        :param Exception exc_value: exception instance that has been
                                    occurred during current context.

        :param traceback traceback: traceback of occurred exception.
        """

        suppressed = isinstance(exc_value, self._errors)
        if suppressed is True and self._log is True:
            logging_services.exception(str(exc_value))

        return suppressed
